import datetime
import importlib
import importlib.util
import inspect
import logging
import os
import pickle
import re
import shutil
import traceback
import zipfile
from copy import deepcopy
from typing import Optional, List, Union
import warnings

import dask
import numpy as np
import pandas as pd
from bson.objectid import ObjectId
from dask.distributed import Client
from sklearn.base import BaseEstimator
from sklearn.dummy import DummyClassifier, DummyRegressor
import joblib
from sklearn.model_selection._split import BaseCrossValidator, BaseShuffleSplit, _RepeatedSplits
from sklearn.inspection import permutation_importance
from photonai.__init__ import __version__
from photonai.base.cache_manager import CacheManager
from photonai.base.photon_elements import Stack, Switch, Preprocessing, CallbackElement, Branch, PipelineElement, \
    PhotonNative
from photonai.base.photon_pipeline import PhotonPipeline
from photonai.base.json_transformer import JsonTransformer
from photonai.helper.helper import PhotonDataHelper
from photonai.optimization import FloatRange
from photonai.photonlogger.logger import logger
from photonai.processing import ResultsHandler
from photonai.optimization.optimization_info import Optimization
from photonai.processing.metrics import Scorer
from photonai.processing.outer_folds import OuterFoldManager
from photonai.processing.photon_folds import FoldInfo
from photonai.processing.results_structure import MDBHyperpipe, MDBHyperpipeInfo, MDBDummyResults, MDBHelper, \
     MDBConfig, MDBOuterFold


class OutputSettings:
    """
    Configuration class that specifies the format in which
    the results are saved. Results can be saved to a MongoDB
    or a simple son-file. You can also choose whether to save
    predictions and/or feature importances.
    """
    def __init__(self,
                 mongodb_connect_url: str = None,
                 save_output: bool = True,
                 overwrite_results: bool = False,
                 generate_best_model: bool = True,
                 user_id: str = '',
                 wizard_object_id: str = '',
                 wizard_project_name: str = '',
                 project_folder: str = ''):
        """
        Initialize the object.

        Parameters:
            mongodb_connect_url:
                Valid mongodb connection url that specifies a database for storing the results.

            save_output:
                Controls the general saving of the results.

            overwrite_results:
                Allows overwriting the results folder if it already exists.

            generate_best_model:
                Determines whether an optimum_pipe should be created and fitted.
                If False, no dependent files are created.

            user_id:
               The user name of the according PHOTONAI Wizard login.

            wizard_object_id:
               The object id to map the designed pipeline in the PHOTONAI Wizard
               to the results in the PHOTONAI CORE Database.

            wizard_project_name:
                How the project is titled in the PHOTONAI Wizard.

            project_folder:
                Deprecated Parameter - transferred to Hyperpipe.

        """
        if project_folder:
            msg = "Deprecated: The parameter 'project_folder' was moved to the Hyperpipe. " \
                  "Please use Hyperpipe(..., project_folder='')."
            logger.error(msg)
            raise DeprecationWarning(msg)
        self.mongodb_connect_url = mongodb_connect_url
        self.overwrite_results = overwrite_results

        self.user_id = user_id
        self.wizard_object_id = wizard_object_id
        self.wizard_project_name = wizard_project_name

        self.generate_best_model = generate_best_model
        self.save_output = save_output
        self.save_predictions_from_best_config_inner_folds = None

        self.verbosity = 0
        self.results_folder = ''
        self.project_folder = ''
        self.log_file = ''
        self.logging_file_handler = None

    # this is only allowed from hyperpipe
    def set_project_folder(self, project_folder):
        self.project_folder = project_folder
        self.initialize_log_file()

    @property
    def setup_error_file(self):
        if self.project_folder:
            return os.path.join(self.project_folder, 'photon_setup_errors.log')
        else:
            return ""

    def initialize_log_file(self):
        self.log_file = self.setup_error_file

    def update_settings(self, name, timestamp):

        if self.save_output:
            if not os.path.exists(self.project_folder):
                os.makedirs(self.project_folder)

            # Todo: give rights to user if this is done by docker container
            if self.overwrite_results:
                self.results_folder = os.path.join(self.project_folder, name + '_results')
            else:
                self.results_folder = os.path.join(self.project_folder, name + '_results_' + timestamp)

            logger.info("Output Folder: " + self.results_folder)

            if not os.path.exists(self.results_folder):
                os.makedirs(self.results_folder)

            if os.path.basename(self.log_file) == "photon_setup_errors.log":
                self.log_file = 'photon_output.log'
            self.log_file = self._add_timestamp(self.log_file)
            self.set_log_file()

        # if we made it here, there should be no further setup errors, every error that comes
        # now can go to the standard logger instance
        if os.path.isfile(self.setup_error_file):
            os.remove(self.setup_error_file)

    def _add_timestamp(self, file):
        return os.path.join(self.results_folder, os.path.basename(file))

    def _get_log_level(self):
        if self.verbosity == 0:
            level = 25
        elif self.verbosity == 1:
            level = logging.INFO  # 20
        elif self.verbosity == 2:
            level = logging.DEBUG  # 10
        else:
            level = logging.WARN  # 30
        return level

    def set_log_file(self):
        logfile_directory = os.path.dirname(self.log_file)
        if not os.path.exists(logfile_directory):
            os.makedirs(logfile_directory)
        if self.logging_file_handler is None:
            self.logging_file_handler = logging.FileHandler(self.log_file)
            self.logging_file_handler.setLevel(self._get_log_level())
            logger.addHandler(self.logging_file_handler)
        else:
            self.logging_file_handler.close()
            self.logging_file_handler.baseFilename = self.log_file

    def set_log_level(self):
        verbose_num = self._get_log_level()
        logger.setLevel(verbose_num)
        for handler in logger.handlers:
            handler.setLevel(verbose_num)


class Hyperpipe(BaseEstimator):
    """The PHOTONAI Hyperpipe class creates a custom
    machine learning pipeline. In addition it defines
    the relevant analysis’ parameters such as the
    cross-validation scheme, the hyperparameter optimization
    strategy, and the performance metrics of interest.

    So called PHOTONAI PipelineElements can be added to
    the Hyperpipe, each of them being a data-processing
    method or a learning algorithm. By choosing and
    combining data-processing methods or algorithms,
    and arranging them with the PHOTONAI classes,
    simple and complex pipeline architectures can be designed rapidly.

    The PHOTONAI Hyperpipe automatizes the nested training,
    test and hyperparameter optimization procedures.

    The Hyperpipe monitors:

    - the nested-cross-validated training
        and test procedure,
    - communicates with the hyperparameter optimization
        strategy,
    - streams information between the pipeline elements,
    - logs all results obtained and evaluates the performance,
    - guides the hyperparameter optimization process by
        a so-called best config metric which is used to select
        the best performing hyperparameter configuration.

    Attributes:
        optimum_pipe (PhotonPipeline):
            An sklearn pipeline object that is fitted to the training data
            according to the best hyperparameter configuration found.
            Currently, we don't create an ensemble of all best hyperparameter
            configs over all folds. We find the best config by comparing
            the test error across outer folds. The hyperparameter config of the best
            fold is used as the optimal model and is then trained on the complete set.

        best_config (dict):
            Dictionary containing the hyperparameters of the
            best configuration. Contains the parameters in the sklearn
            interface of model_name__parameter_name: parameter value.

        results (MDBHyperpipe):
            Object containing all information about the for the
            performed hyperparameter search. Holds the training and test
            metrics for all outer folds, inner folds
            and configurations, as well as additional information.

        elements (list):
            Contains `all PipelineElement or Hyperpipe
            objects that are added to the pipeline.

    Example:
        ``` python
        from photonai.base import Hyperpipe, PipelineElement
        from photonai.optimization import FloatRange
        from sklearn.model_selection import ShuffleSplit, KFold
        from sklearn.datasets import load_breast_cancer

        hyperpipe = Hyperpipe('myPipe',
                              optimizer='random_grid_search',
                              optimizer_params={'limit_in_minutes': 5},
                              outer_cv=ShuffleSplit(test_size=0.2, n_splits=3),
                              inner_cv=KFold(n_splits=10, shuffle=True),
                              metrics=['accuracy', 'precision', 'recall', "f1_score"],
                              best_config_metric='accuracy',
                              eval_final_performance=True,
                              verbosity=0)

        hyperpipe += PipelineElement("SVC", hyperparameters={"C": FloatRange(1, 100)})

        X, y = load_breast_cancer(return_X_y=True)
        hyperpipe.fit(X, y)
        ```

    """
    def __init__(self, name: Optional[str],
                 inner_cv: Union[BaseCrossValidator, BaseShuffleSplit, _RepeatedSplits] = None,
                 outer_cv: Union[BaseCrossValidator, BaseShuffleSplit, _RepeatedSplits, None] = None,
                 optimizer: str = 'grid_search',
                 optimizer_params: dict = None,
                 metrics: Optional[List[Union[Scorer.Metric_Type, str]]] = None,
                 best_config_metric: Optional[Union[Scorer.Metric_Type, str]] = None,
                 eval_final_performance: bool = None,
                 use_test_set: bool = True,
                 test_size: float = 0.2,
                 project_folder: str = '',
                 calculate_metrics_per_fold: bool = True,
                 calculate_metrics_across_folds: bool = False,
                 random_seed: int = None,
                 verbosity: int = 0,
                 learning_curves: bool = False,
                 learning_curves_cut: FloatRange = None,
                 output_settings: OutputSettings = None,
                 performance_constraints: list = None,
                 permutation_id: str = None,
                 cache_folder: str = None,
                 nr_of_processes: int = 1,
                 allow_multidim_targets: bool = False):
        """
        Initialize the object.

        Parameters:
            name:
                Name of hyperpipe instance.

            inner_cv:
                Cross validation strategy to test hyperparameter configurations, generates the validation set.

            outer_cv:
                Cross validation strategy to use for the hyperparameter search itself, generates the test set.

            optimizer:
                Hyperparameter optimization algorithm.

                - In case a string literal is given:
                    - "grid_search": Optimizer that iteratively tests all possible hyperparameter combinations.
                    - "random_grid_search": A variation of the grid search optimization that randomly picks
                        hyperparameter combinations from all possible hyperparameter combinations.
                    - "sk_opt": Scikit-Optimize based on theories of Baysian optimization.
                    - "random_search": randomly chooses hyperparameter from grid-free domain.
                    - "smac": SMAC based on theories of Baysian optimization.
                    - "nevergrad": Nevergrad based on theories of evolutionary learning.

                - In case an object is given:
                    expects the object to have the following methods:
                    - `ask`: returns a hyperparameter configuration in form of an dictionary containing
                        key->value pairs in the sklearn parameter encoding `model_name__parameter_name: parameter_value`
                    - `prepare`: takes a list of pipeline elements and their particular hyperparameters to prepare the
                                 hyperparameter space
                    - `tell`: gets a tested config and the respective performance in order to
                        calculate a smart next configuration to process

            metrics:
                Metrics that should be calculated for both training, validation and test set
                Use the preimported metrics from sklearn and photonai, or register your own

                - Metrics for `classification`:
                    - `accuracy`: sklearn.metrics.accuracy_score
                    - `matthews_corrcoef`: sklearn.metrics.matthews_corrcoef
                    - `confusion_matrix`: sklearn.metrics.confusion_matrix,
                    - `f1_score`: sklearn.metrics.f1_score
                    - `hamming_loss`: sklearn.metrics.hamming_loss
                    - `log_loss`: sklearn.metrics.log_loss
                    - `precision`: sklearn.metrics.precision_score
                    - `recall`: sklearn.metrics.recall_score
                - Metrics for `regression`:
                    - `mean_squared_error`: sklearn.metrics.mean_squared_error
                    - `mean_absolute_error`: sklearn.metrics.mean_absolute_error
                    - `explained_variance`: sklearn.metrics.explained_variance_score
                    - `r2`: sklearn.metrics.r2_score
                - Other metrics
                    - `pearson_correlation`: photon_core.framework.Metrics.pearson_correlation
                    - `variance_explained`:  photon_core.framework.Metrics.variance_explained_score
                    - `categorical_accuracy`: photon_core.framework.Metrics.categorical_accuracy_score

            best_config_metric:
                The metric that should be maximized or minimized in order to choose
                the best hyperparameter configuration.

            eval_final_performance [bool, default=True]:
                DEPRECATED! Use "use_test_set" instead!

            use_test_set [bool, default=True]:
                If the metrics should be calculated for the test set,
                otherwise the test set is seperated but not used.

            project_folder:
                The output folder in which all files generated by the
                PHOTONAI project are saved to.

            test_size:
                The amount of the data that should be left out if no outer_cv is given and
                eval_final_perfomance is set to True.

            calculate_metrics_per_fold:
                If True, the metrics are calculated for each inner_fold.
                If False, calculate_metrics_across_folds must be True.

            calculate_metrics_across_folds:
                If True, the metrics are calculated across all inner_fold.
                If False, calculate_metrics_per_fold must be True.

            random_seed:
                Random Seed.

            verbosity:
                The level of verbosity, 0 is least talkative and
                gives only warn and error, 1 gives adds info and 2 adds debug.

            learning_curves:
                Enables larning curve procedure. Evaluate learning process over
                different sizes of input. Depends on learning_curves_cut.

            learning_curves_cut:
                The tested relativ cuts for data size.

            performance_constraints:
                Objects that indicate whether a configuration should
                be tested further. For example, the inner fold of a config
                does not perform better than the dummy performance.

            permutation_id:
                String identifier for permutation tests.

            cache_folder:
                Folder path for multi-processing.

            nr_of_processes:
                Determined the amount of simultaneous calculation of outer_folds.

            allow_multidim_targets:
                Allows multidimensional targets.

        """

        self.name = re.sub(r'\W+', '', name)

        if eval_final_performance is not None:
            depr_warning = "Hyperpipe parameter eval_final_performance is deprecated. It's called use_test_set now."
            use_test_set = eval_final_performance
            logger.warning(depr_warning)
            raise DeprecationWarning(depr_warning)

        # ====================== Cross Validation ===========================
        # check if both calculate_metrics_per_folds and calculate_metrics_across_folds is False
        if not calculate_metrics_across_folds and not calculate_metrics_per_fold:
            raise NotImplementedError("Apparently, you've set calculate_metrics_across_folds=False and "
                                      "calculate_metrics_per_fold=False. In this case PHOTONAI does not calculate "
                                      "any metrics which doesn't make any sense. Set at least one to True.")
        if inner_cv is None:
            msg = "PHOTONAI requires an inner_cv split. Please enable inner cross-validation. " \
                  "As exmaple: Hyperpipe(...inner_cv = KFold(n_splits = 3), ...). " \
                  "Ensure you import the cross_validation object first."
            logger.error(msg)
            raise AttributeError(msg)

        # use default cut 'FloatRange(0, 1, 'range', 0.2)' if learning_curves = True but learning_curves_cut is None
        if learning_curves and learning_curves_cut is None:
            learning_curves_cut = FloatRange(0, 1, 'range', 0.2)
        elif not learning_curves and learning_curves_cut is not None:
            learning_curves_cut = None

        self.cross_validation = Hyperpipe.CrossValidation(inner_cv=inner_cv,
                                                          outer_cv=outer_cv,
                                                          use_test_set=use_test_set,
                                                          test_size=test_size,
                                                          calculate_metrics_per_fold=calculate_metrics_per_fold,
                                                          calculate_metrics_across_folds=calculate_metrics_across_folds,
                                                          learning_curves=learning_curves,
                                                          learning_curves_cut=learning_curves_cut)

        # ====================== Data ===========================
        self.data = Hyperpipe.Data()

        # ====================== Output Folder and Log File Management ===========================
        if output_settings:
            self.output_settings = output_settings
        else:
            self.output_settings = OutputSettings()

        if project_folder == '':
            self.project_folder = os.getcwd()
        else:
            self.project_folder = project_folder

        self.output_settings.set_project_folder(self.project_folder)

        # update output options to add pipe name and timestamp to results folder
        self._verbosity = 0
        self.verbosity = verbosity
        self.output_settings.set_log_file()

        # ====================== Result Logging ===========================
        self.results_handler = None
        self.results = None
        self.best_config = None

        # ====================== Pipeline ===========================
        self.elements = []
        self._pipe = None
        self.optimum_pipe = None
        self.preprocessing = None

        # ====================== Performance Optimization ===========================
        if optimizer_params is None:
            optimizer_params = {}
        self.optimization = Optimization(metrics=metrics,
                                         best_config_metric=best_config_metric,
                                         optimizer_input=optimizer,
                                         optimizer_params=optimizer_params,
                                         performance_constraints=performance_constraints)

        # self.optimization.sanity_check_metrics()

        # ====================== Caching and Parallelization ===========================
        self.nr_of_processes = nr_of_processes
        if cache_folder:
            self.cache_folder = os.path.join(cache_folder, self.name)
        else:
            self.cache_folder = None

        # ====================== Internals ===========================

        self.permutation_id = permutation_id
        self.allow_multidim_targets = allow_multidim_targets
        self.is_final_fit = False

        # ====================== Random Seed ===========================
        self.random_state = random_seed
        if random_seed is not None:
            import random
            random.seed(random_seed)

    # ===================================================================
    # Helper Classes
    # ===================================================================

    class CrossValidation:

        def __init__(self, inner_cv, outer_cv,
                     use_test_set, test_size,
                     calculate_metrics_per_fold,
                     calculate_metrics_across_folds,
                     learning_curves,
                     learning_curves_cut):
            self.inner_cv = inner_cv
            self.outer_cv = outer_cv
            self.use_test_set = use_test_set
            self.test_size = test_size

            self.learning_curves = learning_curves
            self.learning_curves_cut = learning_curves_cut

            self.calculate_metrics_per_fold = calculate_metrics_per_fold
            # Todo: if self.outer_cv is LeaveOneOut: Set calculate metrics across folds to True -> Print
            self.calculate_metrics_across_folds = calculate_metrics_across_folds

            self.outer_folds = None
            self.inner_folds = dict()

    def __str__(self):
        return "Hyperpipe {}".format(self.name)

    class Data:

        def __init__(self, X=None, y=None, kwargs=None, allow_multidim_targets=False):
            self.X = X
            self.y = y
            self.kwargs = kwargs
            self.allow_multidim_targets = allow_multidim_targets

        def input_data_sanity_checks(self, data, targets, **kwargs):
            # ==================== SANITY CHECKS ===============================
            # 1. Make to numpy arrays
            # 2. erase all Nan targets

            logger.info("Checking input data...")
            self.X = data
            self.y = targets
            self.kwargs = kwargs

            try:
                if self.X is None:
                    raise ValueError("(Input-)data is a NoneType.")
                if self.y is None:
                    raise ValueError("(Input-)target is a NoneType.")

                shape_x = np.shape(self.X)
                shape_y = np.shape(self.y)
                if not self.allow_multidim_targets:
                    if len(shape_y) != 1:
                        if len(np.shape(np.squeeze(self.y))) == 1:
                            # use np.squeeze for non 1D targets.
                            self.y = np.squeeze(self.y)
                            shape_y = np.shape(self.y)
                            msg = "y has been automatically squeezed. If this is not your intention, block this " \
                                  "with Hyperpipe(allow_multidim_targets = True"
                            logger.warning(msg)
                            warnings.warn(msg)
                        else:
                            raise ValueError(
                                "Target is not one-dimensional. Multidimensional targets can cause problems"
                                "with sklearn metrics. Please override with "
                                "Hyperpipe(allow_multidim_targets = True).")
                if not shape_x[0] == shape_y[0]:
                    raise IndexError(
                        "Size of targets mismatch to size of the data: " + str(shape_x[0]) + " - " + str(shape_y[0]))
            except IndexError as ie:
                logger.error("IndexError: " + str(ie))
                raise ie
            except ValueError as ve:
                logger.error("ValueError: " + str(ve))
                raise ve
            except Exception as e:
                logger.error("Error: " + str(e))
                raise e

            # be compatible to list of (image-) files
            if isinstance(self.X, list):
                self.X = np.asarray(self.X)
            elif isinstance(self.X, (pd.DataFrame, pd.Series)):
                self.X = self.X.to_numpy()
            if isinstance(self.y, list):
                self.y = np.asarray(self.y)
            elif isinstance(self.y, pd.Series) or isinstance(self.y, pd.DataFrame):
                self.y = self.y.to_numpy()

            # at first first, erase all rows where y is Nan if preprocessing has not done it already
            try:
                nans_in_y = np.isnan(self.y)
                nr_of_nans = len(np.where(nans_in_y == 1)[0])
                if nr_of_nans > 0:
                    logger.info("You have {} Nans in your target vector, "
                                "PHOTONAI erases every data item that has a Nan Target".format(str(nr_of_nans)))
                    self.X = self.X[~nans_in_y]
                    self.y = self.y[~nans_in_y]
            except Exception as e:
                # This is only for convenience so if it fails then never mind
                logger.error("Removing Nans in target vector failed: " + str(e))
                pass

            logger.info("Running analysis with " + str(self.y.shape[0]) + " samples.")

    # ===================================================================
    # Properties and Helper
    # ===================================================================
    @property
    def estimation_type(self):
        estimation_type = getattr(self.elements[-1], '_estimator_type')
        if estimation_type is None:
            raise NotImplementedError("Last element in Hyperpipe should be an estimator.")
        else:
            return estimation_type

    @property
    def verbosity(self):
        return self._verbosity

    @verbosity.setter
    def verbosity(self, value):
        self._verbosity = value
        self.output_settings.verbosity = self._verbosity
        self.output_settings.set_log_level()

    @staticmethod
    def disable_multiprocessing_recursively(pipe):
        if isinstance(pipe, (Stack, Branch, Switch, Preprocessing)):
            if hasattr(pipe, 'nr_of_processes'):
                pipe.nr_of_processes = 1
            for child in pipe.elements:
                if hasattr(child, 'base_element'):
                    Hyperpipe.disable_multiprocessing_recursively(child.base_element)
        elif isinstance(pipe, PhotonPipeline):
            for name, child in pipe.named_steps.items():
                Hyperpipe.disable_multiprocessing_recursively(child)
        else:
            if hasattr(pipe, 'nr_of_processes'):
                pipe.nr_of_processes = 1

    @staticmethod
    def recursive_cache_folder_propagation(element, cache_folder, inner_fold_id):
        if isinstance(element, (Switch, Stack, Preprocessing)):
            for child in element.elements:
                Hyperpipe.recursive_cache_folder_propagation(child, cache_folder, inner_fold_id)

        elif isinstance(element, Branch):
            # in case it's a Branch, we create a cache subfolder and propagate it to every child
            if cache_folder:
                cache_folder = os.path.join(cache_folder, element.name)
            Hyperpipe.recursive_cache_folder_propagation(element.base_element, cache_folder, inner_fold_id)
            # Hyperpipe.prepare_caching(element.base_element.cache_folder)

        elif isinstance(element, PhotonPipeline):
            element.fold_id = inner_fold_id
            element.cache_folder = cache_folder

            # pipe.caching is automatically set to True or False by .cache_folder setter

            for name, child in element.named_steps.items():
                # we need to check if any element is Branch, Stack or Swtich
                Hyperpipe.recursive_cache_folder_propagation(child, cache_folder, inner_fold_id)

        # else: if it's a simple PipelineElement, then we just don't do anything

    # ===================================================================
    # Pipeline Setup
    # ===================================================================

    def __iadd__(self, pipe_element: PipelineElement):
        """
        Add an element to the machine learning pipeline.
        Returns self.

        Parameters:
            pipe_element:
                The object to add to the machine learning pipeline,
                being either a transformer or an estimator.

        """
        if isinstance(pipe_element, Preprocessing):
            self.preprocessing = pipe_element
        elif isinstance(pipe_element, CallbackElement):
            pipe_element.needs_y = True
            self.elements.append(pipe_element)
        else:
            if isinstance(pipe_element, PipelineElement) or issubclass(type(pipe_element), PhotonNative):
                self.elements.append(pipe_element)
            else:
                raise TypeError("Element must be of type Pipeline Element")
        return self

    def add(self, pipe_element: PipelineElement):
        """
        Add an element to the machine learning pipeline.
        Returns self.

        Parameters:
            pipe_element:
                The object to add to the machine learning pipeline,
                being either a transformer or an estimator.

        """
        self.__iadd__(pipe_element)

    # ===================================================================
    # Workflow Setup
    # ===================================================================
    def _prepare_dummy_estimator(self):
        self.results.dummy_estimator = MDBDummyResults()

        if self.estimation_type == 'regressor':
            self.results.dummy_estimator.strategy = 'mean'
            return DummyRegressor(strategy=self.results.dummy_estimator.strategy)
        elif self.estimation_type == 'classifier':
            self.results.dummy_estimator.strategy = 'most_frequent'
            return DummyClassifier(strategy=self.results.dummy_estimator.strategy)
        else:
            logger.info('Estimator does not specify whether it is a regressor or classifier. '
                        'DummyEstimator step skipped.')
            return

    def __get_pipeline_structure(self, pipeline_elements):
        element_list = dict()
        for p_el in pipeline_elements:
            if not hasattr(p_el, 'name'):
                raise Warning('Strange Pipeline Element found that has no name..? Type: '.format(type(p_el)))
            if hasattr(p_el, 'elements'):
                child_list = self.__get_pipeline_structure(p_el.elements)
                identifier = p_el.name
                if hasattr(p_el, "identifier"):
                    identifier = p_el.identifier + identifier
                    element_list[identifier] = child_list
            else:
                if hasattr(p_el, 'base_element'):
                    element_list[p_el.name] = str(type(p_el.base_element))
                else:
                    element_list[p_el.name] = str(type(p_el))
        return element_list

    def _prepare_result_logging(self, start_time):

        self.results = MDBHyperpipe(name=self.name, version=__version__)
        self.results.hyperpipe_info = MDBHyperpipeInfo()

        # in case eval final performance is false, we have no outer fold predictions
        if not self.cross_validation.use_test_set:
            self.output_settings.save_predictions_from_best_config_inner_folds = True
        self.results_handler = ResultsHandler(self.results, self.output_settings)

        self.results.computation_start_time = start_time
        self.results.hyperpipe_info.estimation_type = self.estimation_type
        self.results.output_folder = self.output_settings.results_folder

        if self.permutation_id is not None:
            self.results.permutation_id = self.permutation_id

        # save wizard information to PHOTONAI db in order to map results to the wizard design object
        if self.output_settings and hasattr(self.output_settings, 'wizard_object_id'):
            if self.output_settings.wizard_object_id:
                self.name = self.output_settings.wizard_object_id
                self.results.name = self.output_settings.wizard_object_id
                self.results.wizard_object_id = ObjectId(self.output_settings.wizard_object_id)
                self.results.wizard_system_name = self.output_settings.wizard_project_name
                self.results.user_id = self.output_settings.user_id
        self.results.outer_folds = []
        self.results.hyperpipe_info.elements = self.__get_pipeline_structure(self.elements)
        self.results.hyperpipe_info.eval_final_performance = self.cross_validation.use_test_set
        self.results.hyperpipe_info.best_config_metric = self.optimization.best_config_metric
        self.results.hyperpipe_info.metrics = self.optimization.metrics
        self.results.hyperpipe_info.learning_curves_cut = self.cross_validation.learning_curves_cut
        self.results.hyperpipe_info.maximize_best_config_metric = self.optimization.maximize_metric

        # optimization
        def _format_cross_validation(cv):
            if cv:
                string = "{}(".format(cv.__class__.__name__)
                for key, val in cv.__dict__.items():
                    string += "{}={}, ".format(key, val)
                return string[:-2] + ")"
            else:
                return "None"

        self.results.hyperpipe_info.cross_validation = \
            {'OuterCV': _format_cross_validation(self.cross_validation.outer_cv),
             'InnerCV': _format_cross_validation(self.cross_validation.inner_cv)}
        self.results.hyperpipe_info.data = {'X_shape': self.data.X.shape, 'y_shape': self.data.y.shape}
        self.results.hyperpipe_info.optimization = {'Optimizer': self.optimization.optimizer_input_str,
                                                    'OptimizerParams': str(self.optimization.optimizer_params),
                                                    'BestConfigMetric': self.optimization.best_config_metric}

        # add json file of hyperpipe attributes
        try:
            json_transformer = JsonTransformer()
            json_transformer.to_json_file(self, self.output_settings.results_folder+"/hyperpipe_config.json")
        except:
            msg = "JsonTransformer was unable to create the .json file."
            logger.warning(msg)
            warnings.warn(msg)

    def _finalize_optimization(self):
        # ==================== EVALUATING RESULTS OF HYPERPARAMETER OPTIMIZATION ===============================
        # 1. computing average metrics
        # 2. finding overall best config
        # 3. training model with best config
        # 4. persisting best model
        logger.clean_info('')
        logger.stars()
        logger.photon_system_log("Finished all outer fold computations.")
        logger.info("Now analysing the final results...")

        # computer dummy metrics
        logger.info("Computing dummy metrics...")
        config_item = MDBConfig()
        dummy_results = [outer_fold.dummy_results for outer_fold in self.results.outer_folds]
        config_item.inner_folds = [f for f in dummy_results if f is not None]
        if len(config_item.inner_folds) > 0:
            self.results.dummy_estimator.metrics_train, self.results.dummy_estimator.metrics_test = \
                MDBHelper.aggregate_metrics_for_inner_folds(config_item.inner_folds, self.optimization.metrics)

        logger.info("Computing mean and std for all outer fold metrics...")
        # Compute all final metrics
        self.results.metrics_train, self.results.metrics_test = \
            MDBHelper.aggregate_metrics_for_outer_folds(self.results.outer_folds, self.optimization.metrics)

        # Find best config across outer folds
        logger.info("Find best config across outer folds...")
        best_config = self.optimization.get_optimum_config_outer_folds(self.results.outer_folds)
        self.best_config = best_config.config_dict
        self.results.best_config = best_config

        # save results again
        self.results.computation_end_time = datetime.datetime.now()
        self.results.computation_completed = True
        logger.info("Save final results...")
        self.results_handler.save()

        logger.info("Prepare Hyperpipe.optimum pipe with best config..")
        # set self to best config
        self.optimum_pipe = self._pipe
        self.optimum_pipe.set_params(**self.best_config)

        if self.output_settings.generate_best_model:
            logger.info("Fitting best model...")
            # set self to best config
            self.optimum_pipe = self._pipe
            self.optimum_pipe.set_params(**self.best_config)

            # set caching
            # we want caching disabled in general but still want to do single subject caching
            self.recursive_cache_folder_propagation(self.optimum_pipe, self.cache_folder, 'fixed_fold_id')
            self.optimum_pipe.caching = False

            # disable multiprocessing when fitting optimum pipe
            # (otherwise inverse_transform won't work for BrainAtlas/Mask)
            self.disable_multiprocessing_recursively(self.optimum_pipe)

            self.optimum_pipe.fit(self.data.X, self.data.y, **self.data.kwargs)

            # Before saving the optimum pipe, add preprocessing without multiprocessing
            self.optimum_pipe.add_preprocessing(self.disable_multiprocessing_recursively(self.preprocessing))

            # Now truly set to no caching (including single_subject_caching)
            self.recursive_cache_folder_propagation(self.optimum_pipe, None, None)

            if self.output_settings.save_output:
                try:
                    pretrained_model_filename = os.path.join(self.output_settings.results_folder,
                                                             'photon_best_model.photon')
                    PhotonModelPersistor.save_optimum_pipe(self.optimum_pipe, pretrained_model_filename)
                    logger.info("Saved best model to file.")
                except Exception as e:
                    logger.info("Could not save best model to file")
                    logger.error(str(e))

                # get feature importances of optimum pipe
                logger.info("Mapping back feature importances...")
                feature_importances = self.optimum_pipe.feature_importances_

                if not feature_importances:
                    logger.info("No feature importances available for {}!".format(self.optimum_pipe.elements[-1][0]))
                else:
                    self.results.best_config_feature_importances = feature_importances

                    # write backmapping file only if optimum_pipes inverse_transform works completely.
                    # restriction: only a faulty inverse_transform is considered, missing ones are further ignored.
                    with warnings.catch_warnings(record=True) as w:
                        # get backmapping
                        backmapping, _, _ = self.optimum_pipe.\
                            inverse_transform(np.array(feature_importances).reshape(1, -1), None)

                        if not any("The inverse transformation is not possible for" in s
                                   for s in [e.message.args[0] for e in w]):
                            # save backmapping
                            self.results_handler.save_backmapping(
                                filename='optimum_pipe_feature_importances_backmapped', backmapping=backmapping)
                        else:
                            logger.info('Could not save feature importance: backmapping NOT successful.')

                # save learning curves
                if self.cross_validation.learning_curves:
                    self.results_handler.save_all_learning_curves()

        logger.info("Summarizing results...")

        logger.info("Write predictions to files...")
        # write all convenience files (summary, predictions_file and plots)
        self.results_handler.write_predictions_file()

        logger.info("Write summary...")
        logger.stars()
        logger.photon_system_log("")
        logger.photon_system_log(self.results_handler.text_summary())

    def preprocess_data(self):
        # if there is a preprocessing pipeline, we apply it first.
        if self.preprocessing is not None:
            logger.info("Applying preprocessing steps...")
            self.preprocessing.fit(self.data.X, self.data.y, **self.data.kwargs)
            self.data.X, self.data.y, self.data.kwargs = self.preprocessing.transform(self.data.X, self.data.y,
                                                                                      **self.data.kwargs)

    def _prepare_pipeline(self):
        self._pipe = Branch.prepare_photon_pipe(self.elements)
        self._pipe = Branch.sanity_check_pipeline(self._pipe)
        if self.random_state:
            self._pipe.random_state = self.random_state

    # ===================================================================
    # sklearn interfaces
    # ===================================================================

    @staticmethod
    def fit_outer_folds(outer_fold_computer, X, y, kwargs, cache_folder):
        try:
            outer_fold_computer.fit(X, y, **kwargs)
        finally:
            CacheManager.clear_cache_files(cache_folder)
        return

    def fit(self, data: np.ndarray, targets: np.ndarray, **kwargs):
        """
        Starts the hyperparameter search and/or fits the pipeline to the data and targets.

        Manages the nested cross validated hyperparameter search:

        1. Filters the data according to filter strategy (1) and according to the imbalanced_data_strategy (2)
        2. requests new configurations from the hyperparameter search strategy, the optimizer,
        3. initializes the testing of a specific configuration,
        4. communicates the result to the optimizer,
        5. repeats 2-4 until optimizer delivers no more configurations to test
        6. finally searches for the best config in all tested configs,
        7. trains the pipeline with the best config and evaluates the performance on the test set

        Parameters:
            data:
                The array-like training and test data with shape=[N, D],
                where N is the number of samples and D is the number of features.

            targets:
                The truth array-like values with shape=[N],
                where N is the number of samples.

            **kwargs:
                Keyword arguments, passed to Outer_Fold_Manager.fit.


        Returns:
            Fitted Hyperpipe.

        """
        # switch to result output folder
        start = datetime.datetime.now()
        self.output_settings.update_settings(self.name, start.strftime("%Y-%m-%d_%H-%M-%S"))

        logger.photon_system_log('=' * 101)
        logger.photon_system_log('PHOTONAI ANALYSIS: ' + self.name)
        logger.photon_system_log('=' * 101)
        logger.info("Preparing data and PHOTONAI objects for analysis...")

        # loop over outer cross validation
        if self.nr_of_processes > 1:
            hyperpipe_client = Client(threads_per_worker=1, n_workers=self.nr_of_processes, processes=False)

        try:
            # check data
            self.data.input_data_sanity_checks(data, targets, **kwargs)
            # create photon pipeline
            self._prepare_pipeline()
            # initialize the progress monitors
            self._prepare_result_logging(start)
            # apply preprocessing
            self.preprocess_data()

            if not self.is_final_fit:

                # Outer Folds
                outer_folds = FoldInfo.generate_folds(self.cross_validation.outer_cv,
                                                      self.data.X, self.data.y, self.data.kwargs,
                                                      self.cross_validation.use_test_set,
                                                      self.cross_validation.test_size)

                self.cross_validation.outer_folds = {f.fold_id: f for f in outer_folds}
                delayed_jobs = []

                # Run Dummy Estimator
                dummy_estimator = self._prepare_dummy_estimator()

                if self.cache_folder is not None:
                    logger.info("Removing cache files...")
                    CacheManager.clear_cache_files(self.cache_folder, force_all=True)

                # loop over outer cross validation
                for i, outer_f in enumerate(outer_folds):

                    # 1. generate OuterFolds Object
                    outer_fold = MDBOuterFold(fold_nr=outer_f.fold_nr)
                    outer_fold_computer = OuterFoldManager(self._pipe,
                                                           self.optimization,
                                                           outer_f.fold_id,
                                                           self.cross_validation,
                                                           cache_folder=self.cache_folder,
                                                           cache_updater=self.recursive_cache_folder_propagation,
                                                           dummy_estimator=dummy_estimator,
                                                           result_obj=outer_fold)
                    # 2. monitor outputs
                    self.results.outer_folds.append(outer_fold)

                    if self.nr_of_processes > 1:
                        result = dask.delayed(Hyperpipe.fit_outer_folds)(outer_fold_computer,
                                                                         self.data.X,
                                                                         self.data.y,
                                                                         self.data.kwargs,
                                                                         self.cache_folder)
                        delayed_jobs.append(result)
                    else:
                        try:
                            # 3. fit
                            outer_fold_computer.fit(self.data.X, self.data.y, **self.data.kwargs)
                            # 4. save outer fold results
                            self.results_handler.save()
                        finally:
                            # 5. clear cache
                            CacheManager.clear_cache_files(self.cache_folder)

                if self.nr_of_processes > 1:
                    dask.compute(*delayed_jobs)
                    self.results_handler.save()

                # evaluate hyperparameter optimization results for best config
                self._finalize_optimization()

                # clear complete cache ?
                CacheManager.clear_cache_files(self.cache_folder, force_all=True)

            ###############################################################################################
            else:
                self.preprocess_data()
                self._pipe.fit(self.data.X, self.data.y, **kwargs)
        except Exception as e:
            logger.error(e)
            logger.error(traceback.format_exc())
            traceback.print_exc()
            raise e
        finally:
            if self.nr_of_processes > 1:
                hyperpipe_client.close()
        return self

    def predict(self, data: np.ndarray, **kwargs) -> np.ndarray:
        """
        Use the optimum pipe to predict the input data.

        Parameters:
            data:
                The array-like prediction data with shape=[M, D],
                where M is the number of samples and D is the number
                of features. D must correspond to the number
                of trained dimensions of the fit method.

            **kwargs:
                Keyword arguments, passed to optimum_pipe.predict.

        Returns:
            Predicted targets calculated on input data with trained model.

        """
        # Todo: if local_search = true then use optimized pipe here?
        if self._pipe:
            return self.optimum_pipe.predict(data, **kwargs)

    def predict_proba(self, data: np.ndarray, **kwargs) -> np.ndarray:
        """
        Use the optimum pipe to predict the probabilities from the input data.

        Parameters:
            data:
                The array-like prediction data with shape=[M, D],
                where M is the number of samples and D is the number
                of features. D must correspond to the number
                of trained dimensions of the fit method.

            **kwargs:
                Keyword arguments, passed to optimum_pipe.predict_proba.

        Returns:
            Probabilities calculated from input data on fitted model.


        """
        if self._pipe:
            return self.optimum_pipe.predict_proba(data, **kwargs)

    def transform(self, data: np.ndarray, **kwargs) -> np.ndarray:
        """
        Use the optimum pipe to transform the data.

        Parameters:
            data:
                The array-like input data with shape=[M, D],
                where M is the number of samples and D is the number
                of features. D must correspond to the number
                of trained dimensions of the fit method.

            **kwargs:
                Keyword arguments, passed to optimum_pipe.transform.

        Returns:
            Transformed data.

        """
        if self._pipe:
            X, _, _ = self.optimum_pipe.transform(data, y=None, **kwargs)
            return X

    def score(self, data: np.ndarray, y: np.ndarray, **kwargs) -> np.ndarray:
        """
        Use the optimum pipe to score the model.

        Parameters:
            data:
                The array-like data with shape=[M, D],
                where M is the number of samples and D is the number
                of features. D must correspond to the number
                of trained dimensions of the fit method.

            y:
                The array-like true targets.

            **kwargs:
                Keyword arguments, passed to optimum_pipe.predict.

        Returns:
            Score data on input data with trained model.

        """
        if self._pipe:
            predictions = self.optimum_pipe.predict(data, **kwargs)
            scorer = Scorer.create(self.optimization.best_config_metric)
            return scorer(y, predictions)

    def _calculate_permutation_importances(self, **kwargs):
        """
        extracted function from get_feature_importance to improve unit testing
        """

        importance_list = {'mean': list(), 'std': list()}

        def train_and_get_fimps(pipeline, train_idx, test_idx, data_X, data_y, data_kwargs, fold_str):

            train_X, train_y, train_kwargs = PhotonDataHelper.split_data(data_X, data_y, data_kwargs,
                                                                         indices=train_idx)

            test_X, test_y, test_kwargs = PhotonDataHelper.split_data(data_X, data_y, data_kwargs,
                                                                      indices=test_idx)

            # fit fold's best model (again) -> to obtain that model's feature importances
            logger.photon_system_log("Permutation Importances: Fitting model for " + fold_str)
            pipeline.fit(train_X, train_y, **train_kwargs)

            # get feature importances
            logger.photon_system_log("Permutation Importances: Calculating performances for " + fold_str)
            perm_imps = permutation_importance(pipeline, test_X, test_y, **kwargs)

            # store into list
            importance_list['mean'].append(perm_imps["importances_mean"])
            importance_list['std'].append(perm_imps["importances_std"])

            return perm_imps

        for outer_fold in self.results.outer_folds:

            if outer_fold.best_config is None:
                raise ValueError("Could not find a best config for outer fold " + str(outer_fold.fold_nr))

            pipe_copy = self.optimum_pipe.copy_me()

            # set pipe to config
            pipe_copy.set_params(**outer_fold.best_config.config_dict)

            if not self.results.hyperpipe_info.eval_final_performance:
                no_outer_cv_indices = False
                if outer_fold.best_config.best_config_score is None:
                    no_outer_cv_indices = True
                if outer_fold.best_config.best_config_score.training is None or not outer_fold.best_config.best_config_score.training.indices:
                    no_outer_cv_indices = True

                if no_outer_cv_indices:
                    data_to_split, y_to_split, kwargs_to_split = self.data.X, self.data.y, self.data.kwargs
                else:

                    logger.photon_system_log("Permutation Importances: Using inner_cv folds.")

                    # get outer fold data
                    idx = outer_fold.best_config.best_config_score.training.indices
                    data_to_split, y_to_split, kwargs_to_split = PhotonDataHelper.split_data(self.data.X,
                                                                                             self.data.y,
                                                                                             self.data.kwargs,
                                                                                             indices=idx)

                for inner_fold in outer_fold.best_config.inner_folds:
                    train_and_get_fimps(pipe_copy,
                                        inner_fold.training.indices, inner_fold.validation.indices,
                                        data_to_split, y_to_split, kwargs_to_split,
                                        "inner fold " + str(inner_fold.fold_nr))

            else:
                train_and_get_fimps(pipe_copy,
                                    outer_fold.best_config.best_config_score.training.indices,
                                    outer_fold.best_config.best_config_score.validation.indices,
                                    self.data.X, self.data.y, self.data.kwargs, "outer fold " + str(outer_fold.fold_nr))

        return importance_list

    def get_permutation_feature_importances(self, **kwargs):
        """
        Fits a model for the best config of each outer fold (using the training data of that fold).
        Then calls sklearn.inspection.permutation_importance with the test data and the given kwargs (e.g. n_repeats).
        Returns mean of "importances_mean" and of "importances_std" of all outer folds.

        Parameters:
            X_val:
                The array-like data with shape=[M, D],
                where M is the number of samples and D is the number
                of features. D must correspond to the number
                of trained dimensions of the fit method.

            y_val:
                The array-like true targets.

            **kwargs:
                Keyword arguments, passed to sklearn.permutation_importance.

        Returns:
            Dictionary with average of "mean" and "std" for all outer folds, respectively.

        """

        logger.photon_system_log("")
        logger.photon_system_log("Computing permutation importances. This may take a while.")
        logger.stars()
        if self.optimum_pipe is None:
            raise ValueError("Cannot calculate permutation importances when optimum_pipe is None (probably the "
                             "training and optimization procedure failed)")
        importance_list = self._calculate_permutation_importances(**kwargs)
        mean_importances = np.mean(np.array(importance_list["mean"]), axis=0)
        std_importances = np.mean(np.array(importance_list["std"]), axis=0)
        logger.stars()

        return {'mean': mean_importances, 'std': std_importances}

    def inverse_transform_pipeline(self, hyperparameters: dict,
                                   data: np.ndarray,
                                   targets: np.ndarray,
                                   data_to_inverse: np.ndarray) -> np.ndarray:
        """
        Inverse transform data for a pipeline with specific hyperparameter configuration.

        1. Copy Sklearn Pipeline,
        2. Set Parameters
        3. Fit Pipeline to data and targets
        4. Inverse transform data with that pipeline

        Parameters:
            hyperparameters:
                The concrete configuration settings for the pipeline elements.

            data:
                The training data to which the pipeline is fitted.

            targets:
                The truth values for training.

            data_to_inverse:
                The data that should be inversed after training.

        Returns:
            Inverse data as array.

        """
        copied_pipe = self.pipe.copy_me()
        copied_pipe.set_params(**hyperparameters)
        copied_pipe.fit(data, targets)
        return copied_pipe.inverse_transform(data_to_inverse)

    # ===================================================================
    # Copy, Save and Load
    # ===================================================================

    def copy_me(self):
        """
        Helper function to copy an entire Hyperpipe

        Returns:
            Hyperpipe

        """
        signature = inspect.getfullargspec(OutputSettings.__init__)[0]
        settings = OutputSettings()
        for attr in signature:
            if hasattr(self.output_settings, attr):
                setattr(settings, attr, getattr(self.output_settings, attr))
        self.output_settings.initialize_log_file()

        # create new Hyperpipe instance
        pipe_copy = Hyperpipe(name=self.name,
                              inner_cv=deepcopy(self.cross_validation.inner_cv),
                              outer_cv=deepcopy(self.cross_validation.outer_cv),
                              best_config_metric=self.optimization.best_config_metric,
                              metrics=self.optimization.metrics,
                              optimizer=self.optimization.optimizer_input_str,
                              optimizer_params=self.optimization.optimizer_params,
                              project_folder=self.project_folder,
                              output_settings=settings)

        signature = inspect.getfullargspec(self.__init__)[0]
        for attr in signature:
            if hasattr(self, attr) and attr != 'output_settings':
                setattr(pipe_copy, attr, getattr(self, attr))

        if hasattr(self, 'preprocessing') and self.preprocessing:
            preprocessing = Preprocessing()
            for element in self.preprocessing.elements:
                preprocessing += element.copy_me()
            pipe_copy += preprocessing
        if hasattr(self, 'elements'):
            for element in self.elements:
                pipe_copy += element.copy_me()
        return pipe_copy

    def save_optimum_pipe(self, filename=None, password=None):
        if filename is None:
            filename = "photon_" + self.name + "_best_model.p"
        PhotonModelPersistor.save_optimum_pipe(self, filename, password)

    @staticmethod
    def load_optimum_pipe(file: str, password: str = None) -> PhotonPipeline:
        """
        Load optimum pipe from file.
        As staticmethod, instantiation is thus not required.
        Called backend: PhotonModelPersistor.load_optimum_pipe.

        Parameters:
            file:
                File path specifying .photon file to load
                trained pipeline from zipped file.

            password:
                Passcode for read file.

        Returns:
            Returns pipeline with all trained PipelineElements.

        """
        return PhotonModelPersistor.load_optimum_pipe(file, password)

    def __repr__(self, **kwargs):
        """Overwrite BaseEstimator's function to avoid errors when using Jupyter Notebooks."""
        return "Hyperpipe(name='{}')".format(self.name)


class PhotonModelPersistor:

    @staticmethod
    def save_elements(elements, folder):
        if not os.path.exists(folder):
            os.makedirs(folder)
        element_identifier = list()

        for i, element in enumerate(elements):
            if hasattr(element, 'disabled'):
                if element.disabled:
                    continue

            # Switch in Switch not covered!
            if isinstance(element, Switch):
                element = element.base_element

            if isinstance(element, (Stack, Branch, Preprocessing)):
                filename = '_' + str(i) + '_' + element.name
                new_folder = os.path.join(folder, filename)
                element_identifier.append({'element_name': element.name, 'filename': filename})
                elements = element.elements
                PhotonModelPersistor.save_elements(elements=elements, folder=new_folder)
                element.elements = []
                joblib.dump(element, os.path.join(folder, filename) + '.pkl', compress=1)
                element_identifier[-1]['mode'] = 'PhotonBuildingBlock'
                element.elements = elements
            else:
                if not hasattr(element, 'base_element'):
                    base_element = element
                else:
                    base_element = element.base_element
                filename = '_' + str(i) + '_' + element.name
                element_identifier.append({'element_name': element.name, 'filename': filename})
                if hasattr(base_element, 'save'):
                    wrapper_file = inspect.getfile(base_element.__class__)
                    base_element.save(os.path.join(folder, filename))
                    element_identifier[-1]['mode'] = 'custom'
                    element_identifier[-1]['wrapper_script'] = os.path.basename(wrapper_file)
                    element_identifier[-1]['test_disabled'] = element.test_disabled
                    element_identifier[-1]['disabled'] = element.disabled
                    element_identifier[-1]['hyperparameters'] = element.hyperparameters
                    # (class_name != element_name) - possibility
                    element_identifier[-1]['class_name'] = type(base_element).__name__
                    shutil.copy(wrapper_file, os.path.join(folder, os.path.basename(wrapper_file)))
                else:
                    try:
                        joblib.dump(element, os.path.join(folder, filename) + '.pkl', compress=1)
                        element_identifier[-1]['mode'] = 'pickle'
                    except:
                        raise NotImplementedError("Custom pipeline element must implement .save() method or "
                                                  "allow pickle.")

        # save pipeline blueprint to make loading of pipeline easier
        with open(os.path.join(folder, '_optimum_pipe_blueprint.pkl'), 'wb') as f:
            pickle.dump(element_identifier, f)

    @staticmethod
    def save_optimum_pipe(optimum_pipe, zip_file: str, password: str = None):
        """
        Save optimal pipeline only. Complete hyperpipe will no not be saved.

        Parameters:
            optimum_pipe:
                The optimum pipe to save.

            zip_file:
                File path as string specifying file to save pipeline to.

            password:
                Password used to encrypt the pipeline file.

        """
        folder = os.path.splitext(zip_file)[0]
        zip_file = folder + '.photon'

        if os.path.exists(folder):
            msg = 'The file you specified already exists as a folder.'
            logger.warning(msg)
            warnings.warn(msg)
        else:
            os.makedirs(folder)

        # only save elements without name. Structure of optimum_pipe.elements: [('name', element),...]
        PhotonModelPersistor.save_elements([val[1] for val in optimum_pipe.elements], folder)

        # write meta infos from pipeline
        with open(os.path.join(folder, '_optimum_pipe_meta.pkl'), 'wb') as f:
            meta_infos = {'photon_version': __version__}
            pickle.dump(meta_infos, f)

        # get all files
        files = list()
        for root, directories, filenames in os.walk(folder):
            for filename in filenames:
                files.append(os.path.join(root, filename))

        if password is not None:
            import pyminizip
            pyminizip.compress(files, zip_file, password)
        else:
            with zipfile.ZipFile(zip_file, 'w') as myzip:
                root_len = len(os.path.dirname(zip_file)) + 1
                for f in files:
                    # in order to work even with subdirectories, we need to substract the dirname from our file
                    # this is why I'm saving the root_len first
                    myzip.write(f, f[root_len:])
                    os.remove(f)
        shutil.rmtree(folder)

    @staticmethod
    def load_elements(folder):
        with open(os.path.join(folder, '_optimum_pipe_blueprint.pkl'), 'rb') as f:
            setup_info = pickle.load(f)
            element_list = list()
            for element_info in setup_info:
                if element_info['mode'] == 'PhotonBuildingBlock':
                    photon_building_block = joblib.load(os.path.join(folder, element_info['filename'] + '.pkl'))
                    base_elements = PhotonModelPersistor.load_elements(os.path.join(folder, element_info['filename']))
                    for _, element in base_elements:
                        photon_building_block += element
                    element_list.append((element_info['element_name'], photon_building_block))
                elif element_info['mode'] == 'custom':
                    if 'class_name' in element_info:  # (class_name != element_name) - possibility
                        spec = importlib.util.spec_from_file_location(element_info['class_name'],
                                                                      os.path.join(folder,
                                                                                   element_info['wrapper_script']))
                        imported_module = importlib.util.module_from_spec(spec)
                        spec.loader.exec_module(imported_module)
                        base_element = getattr(imported_module, element_info['class_name'])
                    else:
                        # backward compatibility
                        try:
                            spec = importlib.util.spec_from_file_location(element_info['element_name'],
                                                                          os.path.join(folder,
                                                                          element_info['wrapper_script']))
                            imported_module = importlib.util.module_from_spec(spec)
                            spec.loader.exec_module(imported_module)
                            base_element = getattr(imported_module, element_info['element_name'])
                        except:
                            msg = "Outdated version: The imported module was created by an older photon version. " \
                                  "Please retrain your model with a newer version."
                            logger.error(msg)
                            raise RuntimeError(msg)
                    custom_element = PipelineElement(name=element_info['element_name'], base_element=base_element(),
                                                     hyperparameters=element_info['hyperparameters'],
                                                     test_disabled=element_info['test_disabled'],
                                                     disabled=element_info['disabled'])
                    custom_element.base_element.load(os.path.join(folder, element_info['filename']))
                    element_list.append((element_info['element_name'], custom_element))
                else:
                    loaded_pipeline_element = joblib.load(os.path.join(folder, element_info['filename'] + '.pkl'))
                    element_list.append((element_info['element_name'], loaded_pipeline_element))
        return element_list

    @staticmethod
    def load_optimum_pipe(file: str, password: str = None):
        """
        Load optimum pipe from file.
        As staticmethod, instantiation is thus not required.
        Called backend: PhotonModelPersistor.load_optimum_pipe.

        Parameters:
            file:
                File path specifying .photon file to load
                trained pipeline from zipped file.

            password:
                Passcode for read file.

        Returns:
            Returns pipeline with all trained PipelineElements.

        """
        if file.endswith('.photon'):
            folder = os.path.dirname(file)
            zf = zipfile.ZipFile(file)
            zf.extractall(folder, pwd=password)
        else:
            raise FileNotFoundError('Specify .photon file that holds PHOTON optimum pipe.')

        load_folder = os.path.join(folder, 'photon_best_model')
        meta_infos = {}
        try:
            with open(os.path.join(load_folder, '_optimum_pipe_meta.pkl'), 'rb') as f:
                meta_infos = pickle.load(f)
        except:
            print("Could not load meta information for optimum pipe")

        element_list = PhotonModelPersistor.load_elements(folder=load_folder)

        # delete unpacked folder to clean up
        # ToDo: Don't unpack at all, but use PHOTON file directly
        from shutil import rmtree
        rmtree(os.path.join(folder, 'photon_best_model'), ignore_errors=True)

        photon_pipe = PhotonPipeline(element_list)
        photon_pipe._meta_information = meta_infos
        return photon_pipe



from sklearn.datasets import load_diabetes
from sklearn.model_selection import KFold, train_test_split

from photonai.base import Hyperpipe, PipelineElement

diabetes = load_diabetes()
X_train, X_val, y_train, y_val = train_test_split(diabetes.data, diabetes.target, random_state=0)

my_pipe = Hyperpipe('basic_ridge_pipe',
                    inner_cv=KFold(n_splits=5),
                    outer_cv=KFold(n_splits=3),
                    optimizer='grid_search',
                    metrics=['mean_absolute_error'],
                    best_config_metric='mean_absolute_error',
                    project_folder='./tmp')

my_pipe += PipelineElement("StandardScaler")
my_pipe += PipelineElement('Ridge', alpha=1e-2)
my_pipe.fit(X_train, y_train)

r = my_pipe.get_permutation_feature_importances(n_repeats=50, random_state=0)

for i in r["mean"].argsort()[::-1]:
    if r["mean"][i] - 2 * r["std"][i] > 0:
        print(f"{diabetes.feature_names[i]:<8}"
              f"{r['mean'][i]:.3f}"
              f" +/- {r['std'][i]:.3f}")



import json
import time
import traceback
import warnings
import datetime
import numpy as np
from typing import Union, List

from photonai.helper.helper import PhotonPrintHelper, PhotonDataHelper, print_double_metrics
from photonai.optimization.performance_constraints import PhotonBaseConstraint
from photonai.photonlogger.logger import logger
from photonai.processing.metrics import Scorer
from photonai.processing.results_structure import MDBHelper, MDBInnerFold, MDBScoreInformation, MDBFoldMetric, MDBConfig

warnings.filterwarnings('ignore', category=DeprecationWarning)
warnings.filterwarnings('ignore', category=FutureWarning)


class InnerFoldManager(object):
    """Inner Fold manager.

    Manages the inner part of the cross validation.

    Information about the inner folds is managed in this object.
    It operates under a specific hyperparameter configuration,
    triggers unified learning, and provides metrics about all inner folds.

    Parameters
    ----------
    pipe_ctor
        The pipeline instance that shall be trained and tested.

    specific_config: dict
        The hyperparameter configuration to test

    optimization_infos: Optimization
        Informations for the optimizer like best_config_metric, maximize_metric, optimizer_params, ...

    cross_validation_infos: CrossValidation
        Infomrations for the inner cross-validation like test_size, eval_final_performance, ...

    outer_fold_id: UUID
        UUID for outer_fold for identification.

    optimization_constraints: PhotonBaseConstraint or List of PhotonBaseConstraint
        Constraints for skipping folds of config if specific constraint occurs.

    raise_error: bool, default=False
        if true, raises exception when training and testing the pipeline fails

    training: bool, default=False
        Mode switch.

    cache_folder: str, default=None
        Path to cache in.

    cache_updater: default=None
        Funtion to update cache.

    """

    def __init__(self, pipe_ctor, specific_config: dict, optimization_infos,
                 cross_validation_infos, outer_fold_id,
                 optimization_constraints: Union[PhotonBaseConstraint, List[PhotonBaseConstraint]] = None,
                 raise_error: bool = False,
                 training: bool = False,
                 cache_folder=None,
                 cache_updater=None):

        self.params = specific_config
        self.pipe = pipe_ctor
        self.optimization_infos = optimization_infos
        self.optimization_constraints = optimization_constraints
        self.outer_fold_id = outer_fold_id
        self.cross_validation_infos = cross_validation_infos

        self.cache_folder = cache_folder
        self.cache_updater = cache_updater

        self.raise_error = raise_error
        self.training = training

    def fit(self, X, y, **kwargs):
        """Iterates over cross-validation folds and trains the pipeline,
        then uses it for predictions. Calculates metrics per fold and averages
        them over fold.

        Parameters
        ----------
        X: ndarray
            Training and test data

        y: ndarray
            Training and test targets

        Returns
        -------
        config_item: MDBConfig
            configuration class for result tree that monitors training and test performance

        """

        config_item = MDBConfig()
        config_item.config_dict = self.params
        config_item.inner_folds = []
        config_item.metrics_test = []
        config_item.metrics_train = []
        config_item.computation_start_time = datetime.datetime.now()

        try:
            # do inner cv
            for idx, (inner_fold_id, inner_fold) in enumerate(self.cross_validation_infos.inner_folds[self.outer_fold_id].items()):

                train, test = inner_fold.train_indices, inner_fold.test_indices

                # split kwargs according to cross validation
                train_X, train_y, kwargs_cv_train = PhotonDataHelper.split_data(X, y, kwargs, indices=train)
                test_X, test_y, kwargs_cv_test = PhotonDataHelper.split_data(X, y, kwargs, indices=test)

                new_pipe = self.pipe()
                if self.cache_folder is not None and self.cache_updater is not None:
                    self.cache_updater(new_pipe, self.cache_folder, inner_fold_id)

                if not config_item.human_readable_config:
                    config_item.human_readable_config = PhotonPrintHelper.config_to_human_readable_dict(new_pipe,
                                                                                                        self.params)
                    logger.clean_info(json.dumps(config_item.human_readable_config, indent=4, sort_keys=True))

                job_data = InnerFoldManager.InnerCVJob(pipe=new_pipe,
                                                       config=dict(self.params),
                                                       metrics=self.optimization_infos.metrics,
                                                       callbacks=self.optimization_constraints,
                                                       train_data=InnerFoldManager.JobData(train_X, train_y, train,
                                                                                           kwargs_cv_train),
                                                       test_data=InnerFoldManager.JobData(test_X, test_y, test,
                                                                                          kwargs_cv_test))

                # only for unparallel processing
                # inform children in which inner fold we are
                # self.pipe.distribute_cv_info_to_hyperpipe_children(inner_fold_counter=fold_cnt)
                # self.mother_inner_fold_handle(fold_cnt)

                # --> write that output in InnerFoldManager!
                # logger.debug(config_item.human_readable_config)
                fold_nr = idx + 1
                logger.debug('calculating inner fold ' + str(fold_nr) + '...')

                curr_test_fold, curr_train_fold = InnerFoldManager.fit_and_score(job_data)
                if self.cross_validation_infos.learning_curves:
                    learning_curves = self.compute_learning_curves(new_pipe, train_X, train_y, train, kwargs_cv_train,
                                                                   test_X, test_y, test, kwargs_cv_test)
                    learning_curves.append([1., curr_test_fold.metrics, curr_train_fold.metrics])
                else:
                    learning_curves = list()

                logger.debug('Performance inner fold ' + str(fold_nr))
                print_double_metrics(curr_train_fold.metrics, curr_test_fold.metrics, photon_system_log=False)

                durations = job_data.pipe.time_monitor

                self.update_config_item_with_inner_fold(config_item=config_item,
                                                        fold_cnt=fold_nr,
                                                        curr_train_fold=curr_train_fold,
                                                        curr_test_fold=curr_test_fold,
                                                        time_monitor=durations,
                                                        feature_importances=new_pipe.feature_importances_,
                                                        learning_curves=learning_curves)

                if isinstance(self.optimization_constraints, list):
                    break_cv = 0
                    for cf in self.optimization_constraints:
                        if not cf.shall_continue(config_item):
                            logger.info('Skipped further cross validation after fold ' + str(fold_nr) +
                                        ' due to performance constraints in ' + cf.metric)
                            break_cv += 1
                            break
                    if break_cv > 0:
                        break
                elif isinstance(self.optimization_constraints, PhotonBaseConstraint):
                    if not self.optimization_constraints.shall_continue(config_item):
                        logger.info('Skipped further cross validation after fold ' + str(fold_nr) +
                                    ' due to performance constraints in ' + self.optimization_constraints.metric)
                        break

            InnerFoldManager.process_fit_results(config_item,
                                                 self.cross_validation_infos.calculate_metrics_across_folds,
                                                 self.cross_validation_infos.calculate_metrics_per_fold,
                                                 self.optimization_infos.metrics)

        except Exception as e:
            if self.raise_error:
                raise e
            logger.error(e)
            logger.error(traceback.format_exc())
            traceback.print_exc()
            if not isinstance(e, Warning):
                config_item.config_failed = True
            config_item.config_error = str(e)
            warnings.warn('One test iteration of pipeline failed with error')

        logger.debug('...done with')
        logger.debug(json.dumps(config_item.human_readable_config, indent=4, sort_keys=True))

        config_item.computation_end_time = datetime.datetime.now()
        return config_item

    def compute_learning_curves(self, new_pipe, train_X, train_y, train, kwargs_cv_train,
                                test_X, test_y, test, kwargs_cv_test):
        self.cross_validation_infos.learning_curves_cut.transform()
        cut_range = [round(cut * train_X.shape[0]) for cut in self.cross_validation_infos.learning_curves_cut.values]
        learning_curves = []
        for i, cut in enumerate(cut_range[1:]):
            cut_indices = np.arange(cut)
            train_cut_X, train_cut_y, train_cut_kwargs = PhotonDataHelper.split_data(train_X, train_y, kwargs_cv_train,
                                                                                     indices=cut_indices)
            train_cut = train[:cut]
            job_data = self.InnerCVJob(pipe=new_pipe,
                                       config=dict(self.params),
                                       metrics=self.optimization_infos.metrics,
                                       callbacks=self.optimization_constraints,
                                       train_data=self.JobData(train_cut_X, train_cut_y, train_cut, train_cut_kwargs),
                                       test_data=self.JobData(test_X, test_y, test, kwargs_cv_test))
            curr_test_cut, curr_train_cut = InnerFoldManager.fit_and_score(job_data)
            learning_curves.append([self.cross_validation_infos.learning_curves_cut.values[i], curr_test_cut.metrics,
                                    curr_train_cut.metrics])
        return learning_curves

    class JobData:
        def __init__(self, X, y, indices, cv_kwargs):
            self.X = X
            self.y = y
            self.indices = indices
            self.cv_kwargs = cv_kwargs

    class InnerCVJob:

        def __init__(self, pipe, config, metrics, callbacks, train_data, test_data):
            self.pipe = pipe
            self.config = config
            self.metrics = metrics
            self.callbacks = callbacks
            self.train_data = train_data
            self.test_data = test_data

    @staticmethod
    def update_config_item_with_inner_fold(config_item, fold_cnt, curr_train_fold, curr_test_fold, time_monitor,
                                           feature_importances, learning_curves):
        # fill result tree with fold information
        inner_fold = MDBInnerFold()
        inner_fold.fold_nr = fold_cnt
        inner_fold.training = curr_train_fold
        inner_fold.validation = curr_test_fold

        inner_fold.number_samples_validation = len(curr_test_fold.indices)
        inner_fold.number_samples_training = len(curr_train_fold.indices)
        inner_fold.time_monitor = time_monitor
        inner_fold.feature_importances = feature_importances
        inner_fold.learning_curves = learning_curves
        # save all inner folds to the tree under the config item
        config_item.inner_folds.append(inner_fold)

    @staticmethod
    def process_fit_results(config_item,
                            calculate_metrics_across_folds,
                            calculate_metrics_per_fold,
                            metrics):

        overall_y_pred_test = []
        overall_y_true_test = []
        overall_y_pred_train = []
        overall_y_true_train = []

        for fold in config_item.inner_folds:
            curr_test_fold = fold.validation
            curr_train_fold = fold.training

            if calculate_metrics_across_folds:
                # if we have one hot encoded values -> concat horizontally
                if isinstance(curr_test_fold.y_pred, np.ndarray):
                    if len(curr_test_fold.y_pred.shape) > 1:
                        axis = 1
                    else:
                        axis = 0
                else:
                    # if we have lists concat
                    axis = 0
                overall_y_true_test = np.concatenate((overall_y_true_test, curr_test_fold.y_true), axis=axis)
                overall_y_pred_test = np.concatenate((overall_y_pred_test, curr_test_fold.y_pred), axis=axis)

                # we assume y_pred from the training set comes in the same shape as y_pred from the test se
                overall_y_true_train = np.concatenate((overall_y_true_train, curr_train_fold.y_true), axis=axis)
                overall_y_pred_train = np.concatenate((overall_y_pred_train, curr_train_fold.y_pred), axis=axis)

                # metrics across folds
                metrics_to_calculate = list(metrics)
                if 'score' in metrics_to_calculate:
                    metrics_to_calculate.remove('score')
                metrics_train = Scorer.calculate_metrics(overall_y_true_train,
                                                         overall_y_pred_train, metrics_to_calculate)
                metrics_test = Scorer.calculate_metrics(overall_y_true_test,
                                                        overall_y_pred_test, metrics_to_calculate)

                def metric_to_db_class(metric_list):
                    db_metrics = []
                    for metric_name, metric_value in metric_list.items():
                        new_metric = MDBFoldMetric(operation="raw", metric_name=metric_name,
                                                   value=metric_value)
                        db_metrics.append(new_metric)
                    return db_metrics

                db_metrics_train = metric_to_db_class(metrics_train)
                db_metrics_test = metric_to_db_class(metrics_test)

                # if we want to have metrics for each fold as well, calculate mean and std.
                if calculate_metrics_per_fold:
                    db_metrics_fold_train, db_metrics_fold_test = MDBHelper.aggregate_metrics_for_inner_folds(config_item.inner_folds,
                                                                                                              metrics)
                    config_item.metrics_train = db_metrics_train + db_metrics_fold_train
                    config_item.metrics_test = db_metrics_test + db_metrics_fold_test
                else:
                    config_item.metrics_train = db_metrics_train
                    config_item.metrics_test = db_metrics_test

            elif calculate_metrics_per_fold:
                # calculate mean and std over all fold metrics
                config_item.metrics_train, config_item.metrics_test = MDBHelper.aggregate_metrics_for_inner_folds(config_item.inner_folds,
                                                                                                                  metrics)

    @staticmethod
    def fit_and_score(job: InnerCVJob):

        pipe = job.pipe

        # set params to current config
        pipe.set_params(**job.config)

        # start fitting
        pipe.fit(job.train_data.X, job.train_data.y, **job.train_data.cv_kwargs)

        logger.debug('Scoring Training Data')

        # score test data
        curr_test_fold = InnerFoldManager.score(pipe, job.test_data.X, job.test_data.y, job.metrics,
                                                indices=job.test_data.indices,
                                                **job.test_data.cv_kwargs)

        logger.debug('Scoring Test Data')
        # score train data
        curr_train_fold = InnerFoldManager.score(pipe, job.train_data.X, job.train_data.y, job.metrics,
                                                 indices=job.train_data.indices,
                                                 training=True, **job.train_data.cv_kwargs)

        return curr_test_fold, curr_train_fold

    @staticmethod
    def score(estimator, X, y_true, metrics, indices=[],
              calculate_metrics: bool=True, training: bool=False, **kwargs):
        """Uses the pipeline to predict the given data,
        compare it to the truth values and calculate metrics

        Parameters
        ----------
        estimator:
            The pipeline or pipeline element for prediction.
            Requires the 'predict' method.


        X: ndarray
            The data for prediction.

        y_true: ndarray
            The truth values for the data.

        metrics: list
            The metrics to be calculated.

        indices: list
            The indices of the given data and targets that are logged into the result tree.

        training: list
            if True, all training_only pipeline elements are executed, if False they are skipped

        calculate_metrics: bool, default=True
            if True, calculates metrics for given data

        training: bool, default=False
            If True, an estimator.transform() is prepended here.


        Returns
        -------
        information: MDBScoreInformation
            Informations about scorer.

        """

        scoring_time_start = time.time()

        output_metrics = {}

        if not training:
            y_pred = estimator.predict(X, **kwargs)
        else:
            X, y_true_new, kwargs_new = estimator.transform(X, y_true, **kwargs)
            if y_true_new is not None:
                y_true = y_true_new
            if kwargs_new is not None and len(kwargs_new) > 0:
                kwargs = kwargs_new
            y_pred = estimator.predict(X, training=True, **kwargs)

        # Nice to have
        # InnerFoldManager.plot_some_data(y_true, y_pred)

        if calculate_metrics:
            if isinstance(y_pred, np.ndarray) and y_pred.dtype.names:
                y_pred_names = [y_pred.dtype.names]
                if "y_pred" not in y_pred_names[0]:
                    msg = "If scorer object does not return 1d array or list, PHOTON expected name 'y_pred' in nd array."
                    logger.error(msg)
                    raise KeyError(msg)
                score_metrics = Scorer.calculate_metrics(y_true, y_pred["y_pred"], metrics)
            else:
                y_pred_names = []
                score_metrics = Scorer.calculate_metrics(y_true, y_pred, metrics)

            # add default metric
            if output_metrics:
                output_metrics = {**output_metrics, **score_metrics}
            else:
                output_metrics = score_metrics
        else:
            output_metrics = {}

        final_scoring_time = time.time() - scoring_time_start

        probabilities = []
        if hasattr(estimator, '_final_estimator'):
            if hasattr(estimator._final_estimator.base_element, 'predict_proba'):
                probabilities = estimator.predict_proba(X, training=training,  **kwargs)

                try:
                    if probabilities is not None:
                        if not len(probabilities) == 0:
                            probabilities = probabilities.tolist()
                except:
                    warnings.warn('No probabilities available.')

        if not isinstance(y_pred, list):
            y_pred = y_pred_names + np.asarray(y_pred).tolist()

        if not isinstance(y_true, list):
            y_true = np.asarray(y_true).tolist()

        score_result_object = MDBScoreInformation(metrics=output_metrics,
                                                  score_duration=final_scoring_time,
                                                  y_pred=y_pred, y_true=y_true,
                                                  indices=np.asarray(indices).tolist(),
                                                  probabilities=probabilities)

        return score_result_object



import datetime
import os

import numpy as np
import warnings
from sklearn.utils.metaestimators import _BaseComposition

from photonai.base.cache_manager import CacheManager
from photonai.helper.helper import PhotonDataHelper
from photonai.photonlogger.logger import logger


class PhotonPipeline(_BaseComposition):

    def __init__(self, elements, random_state=False):
        self.elements = elements
        self.random_state = random_state
        self.current_config = None
        # caching stuff
        self.caching = False
        self._fold_id = None
        self._cache_folder = None
        self.time_monitor = {'fit': [], 'transform_computed': [], 'transform_cached': [], 'predict': []}
        self.cache_man = None

        # helper for single subject caching
        self._single_subject_caching = False
        self._fix_fold_id = False
        self._do_not_delete_cache_folder = False
        self._parallel_use = False

        # helper for optimum pipe
        self._meta_information = None

        # used in parallelization
        self.skip_loading = False


    def set_lock(self, lock):
        self.cache_man.lock = lock

    @property
    def single_subject_caching(self):
        return self._single_subject_caching

    @single_subject_caching.setter
    def single_subject_caching(self, value: bool):
        if value:
            self._fix_fold_id = True
            self._do_not_delete_cache_folder = True
        else:
            self._fix_fold_id = False
            self._do_not_delete_cache_folder = False
        self._single_subject_caching = value

    @property
    def fold_id(self):
        return self._fold_id

    @fold_id.setter
    def fold_id(self, value):
        if value is None:
            self._fold_id = ''
            # we dont need group-wise caching if we have no inner fold id
            self.caching = False
            self.cache_man = None
        else:
            if self._fix_fold_id:
                self._fold_id = "fixed_fold_id"
            else:
                self._fold_id = str(value)
            self.caching = True
            self.cache_man = CacheManager(self._fold_id, self.cache_folder, self._parallel_use,
                                          self._single_subject_caching)

    @property
    def cache_folder(self):
        return self._cache_folder

    @cache_folder.setter
    def cache_folder(self, value):

        if not self._do_not_delete_cache_folder:
            self._cache_folder = value
        else:
            if isinstance(value, str) and not value.endswith("DND"):
                self._cache_folder = value + "DND"
            else:
                self._cache_folder = value

        if isinstance(self._cache_folder, str):
            self.caching = True
            if not os.path.isdir(self._cache_folder):
                os.makedirs(self._cache_folder)
            self.cache_man = CacheManager(self._fold_id, self.cache_folder, self._parallel_use, self._single_subject_caching)
        else:
            self.caching = False

    def get_params(self, deep=True):
        """Get parameters for this estimator.
        Parameters
        ----------
        deep : boolean, optional
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.
        Returns
        -------
        params : mapping of string to any
            Parameter names mapped to their values.
        """
        return self._get_params('elements', deep=deep)

    def set_params(self, **kwargs):
        """Set the parameters of this estimator.
        Valid parameter keys can be listed with ``get_params()``.
        Returns
        -------
        self
        """
        if self.current_config is not None and len(self.current_config) > 0:
            if kwargs is not None and len(kwargs) == 0:
                raise ValueError("Pipeline cannot set parameters to elements with an emtpy dictionary. Old values persist")
        self.current_config = kwargs
        self._set_params('elements', **kwargs)

        return self

    def _validate_elements(self):
        names, estimators = zip(*self.elements)

        # validate names
        self._validate_names(names)

        # validate estimators
        transformers = estimators[:-1]
        estimator = estimators[-1]

        for t in transformers:
            if t is None:
                continue
            if not (hasattr(t, "fit") or not hasattr(t, "transform")):
                raise TypeError("All intermediate elements should be "
                                "transformers and implement fit and transform."
                                " '%s' (type %s) doesn't" % (t, type(t)))

        # We allow last estimator to be None as an identity transformation
        if estimator is not None and not hasattr(estimator, "fit"):
            raise TypeError("Last step of Pipeline should implement fit. "
                            "'%s' (type %s) doesn't"
                            % (estimator, type(estimator)))

    def fit(self, X, y=None, **kwargs):

        self._validate_elements()
        X, y, kwargs = self._caching_fit_transform(X, y, kwargs, fit=True)

        if self._final_estimator is not None:
            logger.debug('PhotonPipeline: Fitting ' + self._final_estimator.name)
            fit_start_time = datetime.datetime.now()
            if self.random_state:
                self._final_estimator.random_state = self.random_state
            self._final_estimator.fit(X, y, **kwargs)
            n = PhotonDataHelper.find_n(X)
            fit_duration = (datetime.datetime.now() - fit_start_time).total_seconds()
            self.time_monitor['fit'].append((self.elements[-1][0], fit_duration, n))
        return self

    def check_for_numpy_array(self, list_object):
        # be compatible to list of (image-) files
        if isinstance(list_object, list):
            return np.asarray(list_object)
        else:
            return list_object

    def transform(self, X, y=None, **kwargs):
        """
        Calls transform on every step that offers a transform function
        including the last step if it has the transformer flag,
        and excluding the last step if it has the estimator flag but no transformer flag.

        Returns transformed X, y and kwargs
        """
        if self.single_subject_caching:
            initial_X = np.array(X)
        else:
            initial_X = None

        X, y, kwargs = self._caching_fit_transform(X, y, kwargs)

        if self._final_estimator is not None:
            if self._estimator_type is None:
                if self.caching:
                    X, y, kwargs = self.load_or_save_cached_data(self._final_estimator.name, X, y, kwargs,
                                                                 self._final_estimator,
                                                                 initial_X=initial_X)
                else:
                    logger.debug('PhotonPipeline: Transforming data with ' + self._final_estimator.name)
                    X, y, kwargs = self._final_estimator.transform(X, y, **kwargs)

        return X, y, kwargs

    def load_or_save_cached_data(self, name, X, y, kwargs, transformer, fit=False,
                                 needed_for_further_computation=False, initial_X=None):
        if not self.single_subject_caching:
            # if we do it group-wise then its easy
            if self.skip_loading and not needed_for_further_computation:
                # check if data is already calculated
                if self.cache_man.check_cache(name):
                    # if so, do nothing
                    return X, y, kwargs
                else:
                    # otherwise, do the calculation and save it
                    cached_result = None
            else:
                start_time_for_loading = datetime.datetime.now()
                cached_result = self.cache_man.load_cached_data(name)

            if cached_result is None:
                X, y, kwargs = self._do_timed_fit_transform(name, transformer, fit, X, y, **kwargs)

                start_time_saving = datetime.datetime.now()
                self.cache_man.save_data_to_cache(name, (X, y, kwargs))
                saving_duration = (datetime.datetime.now() - start_time_saving).total_seconds()
                self.time_monitor['transform_cached'].append((name, saving_duration, 1))
            else:
                X, y, kwargs = cached_result[0], cached_result[1], cached_result[2]
                loading_duration = (datetime.datetime.now() - start_time_for_loading).total_seconds()
                n = PhotonDataHelper.find_n(X)
                self.time_monitor['transform_cached'].append((name, loading_duration, n))
            return X, y, kwargs
        else:
            # if we do it subject-wise we need to iterate and collect the results
            processed_X, processed_y, processed_kwargs = list(), list(), dict()
            X_uncached, y_uncached, kwargs_uncached, initial_X_uncached = list(), list(), dict(), list()
            list_of_idx_cached, list_of_idx_non_cached = list(), list()

            nr = PhotonDataHelper.find_n(X)
            for start, stop in PhotonDataHelper.chunker(nr, 1):
                # split data in single entities, find key from first element = PATH to file
                X_key, _, _ = PhotonDataHelper.split_data(initial_X, None, {}, start, stop)
                X_batched, y_batched, kwargs_dict_batched = PhotonDataHelper.split_data(X, y, kwargs, start, stop)
                self.cache_man.update_single_subject_state_info(X_key)

                # check if item has been processed
                if self.cache_man.check_cache(name):
                    list_of_idx_cached.append(start)
                else:
                    list_of_idx_non_cached.append(start)
                    X_uncached = PhotonDataHelper.stack_data_vertically(X_uncached, X_batched)
                    y_uncached = PhotonDataHelper.stack_data_vertically(y_uncached, y_batched)
                    initial_X_uncached = PhotonDataHelper.stack_data_vertically(initial_X_uncached, X_key)
                    kwargs_uncached = PhotonDataHelper.join_dictionaries(kwargs_uncached, kwargs_dict_batched)

            # now we know which part can be loaded and which part should be transformed
            # first apply the transformation to the group, then save it single-subject-wise
            if len(list_of_idx_non_cached) > 0:

                # apply transformation groupwise
                new_group_X, new_group_y, new_group_kwargs = self._do_timed_fit_transform(name, transformer, fit,
                                                                                          X_uncached,
                                                                                          y_uncached,
                                                                                          **kwargs_uncached)

                # then save it single
                nr = PhotonDataHelper.find_n(new_group_X)
                for start in range(nr):
                    # split data in single entities
                    X_batched, y_batched, kwargs_dict_batched = PhotonDataHelper.split_data(new_group_X,
                                                                                            new_group_y,
                                                                                            new_group_kwargs,
                                                                                            start, start)
                    X_key, _, _ = PhotonDataHelper.split_data(initial_X_uncached, None, {}, start, start)
                    # we save the data in relation to the input path (X_key = hash(input X))
                    self.cache_man.update_single_subject_state_info(X_key)

                    start_time_saving = datetime.datetime.now()
                    self.cache_man.save_data_to_cache(name, (X_batched, y_batched, kwargs_dict_batched))
                    saving_duration = (datetime.datetime.now() - start_time_saving).total_seconds()
                    self.time_monitor['transform_cached'].append((name, saving_duration, 1))

                # we need to collect the data only when we want to load them
                # we can skip that process if we only want them to get into the cache (case: parallelisation)
                if not self.skip_loading or needed_for_further_computation:
                    # stack results
                    processed_X, processed_y, processed_kwargs = new_group_X, new_group_y, new_group_kwargs

            # afterwards load everything that has been cached
            if len(list_of_idx_cached) > 0:
                if not self.skip_loading or needed_for_further_computation:
                    for cache_idx in list_of_idx_cached:
                        # we identify the data according to the input path (X before any transformation)
                        self.cache_man.update_single_subject_state_info([initial_X[cache_idx]])

                        # time the loading of the cached item
                        start_time_for_loading = datetime.datetime.now()
                        transformed_X, transformed_y, transformed_kwargs = self.cache_man.load_cached_data(name)
                        loading_duration = (datetime.datetime.now() - start_time_for_loading).total_seconds()
                        self.time_monitor['transform_cached'].append((name, loading_duration, PhotonDataHelper.find_n(X)))

                        processed_X, processed_y, processed_kwargs = PhotonDataHelper.join_data(processed_X, transformed_X,
                                                                                                processed_y, transformed_y,
                                                                                                processed_kwargs, transformed_kwargs)

            logger.debug(name + " loaded " + str(len(list_of_idx_cached)) + " items from cache and computed "
                         + str(len(list_of_idx_non_cached)))
            if not self.skip_loading or needed_for_further_computation:
                # now sort the data in the correct order again
                processed_X, processed_y, processed_kwargs = PhotonDataHelper.resort_splitted_data(processed_X,
                                                                                                   processed_y,
                                                                                                   processed_kwargs,
                                                                                                   PhotonDataHelper.stack_data_vertically(
                                                                                                       list_of_idx_cached,
                                                                                                       list_of_idx_non_cached))

            return processed_X, processed_y, processed_kwargs

    def _do_timed_fit_transform(self, name, transformer, fit, X, y, **kwargs):

        n = PhotonDataHelper.find_n(X)
        if self.random_state:
            transformer.random_state = self.random_state

        if fit:
            logger.debug('PhotonPipeline: Fitting ' + transformer.name)
            fit_start_time = datetime.datetime.now()
            transformer.fit(X, y, **kwargs)
            fit_duration = (datetime.datetime.now() - fit_start_time).total_seconds()
            self.time_monitor['fit'].append((name, fit_duration, n))

        logger.debug('PhotonPipeline: Transforming data with ' + transformer.name)
        transform_start_time = datetime.datetime.now()
        X, y, kwargs = transformer.transform(X, y, **kwargs)
        transform_duration = (datetime.datetime.now() - transform_start_time).total_seconds()
        self.time_monitor['transform_computed'].append((name, transform_duration, n))
        return X, y, kwargs

    def _caching_fit_transform(self, X, y, kwargs, fit=False):

        if self.single_subject_caching:
            initial_X = np.array(X)
        else:
            initial_X = None

        if self.caching:
            # update infos, just in case
            self.cache_man.hash = self._fold_id
            self.cache_man.cache_folder = self.cache_folder
            if not self.single_subject_caching:
                self.cache_man.prepare([name for name, e in self.elements], self.current_config, X)
            else:
                self.cache_man.prepare([name for name, e in self.elements], self.current_config, single_subject_caching=True)
            last_cached_item = None

        # all elements except the last one
        num_steps = len(self.elements) - 1

        for num, (name, transformer) in enumerate(self.elements[:-1]):
            if not self.caching or self.current_config is None or \
                    (hasattr(transformer, 'skip_caching') and transformer.skip_caching):
                X, y, kwargs = self._do_timed_fit_transform(name, transformer, fit, X, y, **kwargs)
            else:
                # load data when the first item occurs that needs new calculation
                if self.cache_man.check_cache(name):
                    # as long as we find something cached, we remember what it was
                    last_cached_item = name
                    # if it is the last step, we need to load the data now
                    if num + 1 == num_steps and not self.skip_loading:
                        X, y, kwargs = self.load_or_save_cached_data(last_cached_item, X, y, kwargs, transformer, fit,
                                                                     initial_X=initial_X)
                else:
                    if last_cached_item is not None:
                        # we load the cached data when the first transformation on this data is upcoming
                        X, y, kwargs = self.load_or_save_cached_data(last_cached_item, X, y, kwargs, transformer, fit,
                                                                     needed_for_further_computation=True,
                                                                     initial_X=initial_X)
                    X, y, kwargs = self.load_or_save_cached_data(name, X, y, kwargs, transformer, fit,
                                                                 initial_X=initial_X)

            # always work with numpy arrays to avoid checking for shape attribute
            X = self.check_for_numpy_array(X)
            y = self.check_for_numpy_array(y)

        return X, y, kwargs

    def predict(self, X, training=False, **kwargs):
        """
        Transforms the data for every step that offers a transform function
        and then calls the estimator with predict on transformed data.
        It returns the predictions made.

        In case the last step is no estimator, it returns the transformed data.
        """

        # first transform
        if not training:
            X, _, kwargs = self.transform(X, y=None, **kwargs)

        # then call predict on final estimator
        if self._final_estimator is not None:
            if self._final_estimator.is_estimator:
                logger.debug('PhotonPipeline: Predicting with ' + self._final_estimator.name + ' ...')
                predict_start_time = datetime.datetime.now()
                y_pred = self._final_estimator.predict(X, **kwargs)
                predict_duration = (datetime.datetime.now() - predict_start_time).total_seconds()
                n = PhotonDataHelper.find_n(X)
                self.time_monitor['predict'].append((self.elements[-1][0], predict_duration, n))
                return y_pred
            else:
                return X
        else:
            return None

    def predict_proba(self, X, training: bool=False, **kwargs):
        if not training:
            X, _, kwargs = self.transform(X, y=None, **kwargs)

        if self._final_estimator is not None:
            if self._final_estimator.is_estimator:
                if hasattr(self._final_estimator, "predict_proba"):
                    if hasattr(self._final_estimator, 'needs_covariates'):
                        if self._final_estimator.needs_covariates:
                            return self._final_estimator.predict_proba(X, **kwargs)
                        else:
                            return self._final_estimator.predict_proba(X)
                    else:
                        return self._final_estimator.predict_proba(X)

        raise NotImplementedError("The final estimator does not have a predict_proba method")

    def inverse_transform(self, X, y=None, **kwargs):
        # simply use X to apply inverse_transform
        # does not work on any transformers changing y or kwargs!
        for name, transform in reversed(self.elements):
            if transform.disabled:
                continue
            try:
                X, y, kwargs = transform.inverse_transform(X, y, **kwargs)
            except Exception as e:
                msg = "The inverse transformation is not possible for {0}. " \
                      "Cause: {1} " \
                      "The returned value is based on the inverse input data of {0}.".format(name, str(e))
                warnings.warn(msg)
                logger.photon_system_log(msg)
                break

        return X, y, kwargs

    def fit_transform(self, X, y=None, **kwargs):
        # return self.fit(X, y, **kwargs).transform(X, y, **kwargs)
        raise NotImplementedError('fit_transform not yet implemented in PHOTON Pipeline')

    def fit_predict(self, X, y=None, **kwargs):
        raise NotImplementedError('fit_predict not yet implemented in PHOTON Pipeline')

    def copy_me(self):
        pipeline_steps = []
        for item_name, item in self.elements:
            cpy = item.copy_me()
            if isinstance(cpy, list):
                for new_step in cpy:
                    pipeline_steps.append((new_step.name, new_step))
            else:
                pipeline_steps.append((cpy.name, cpy))
        new_pipe = PhotonPipeline(pipeline_steps)
        new_pipe.random_state = self.random_state
        return new_pipe

    def score(self, X: np.ndarray, y: np.ndarray, **kwargs) -> float:
        """
        Transforms the data for every step that offers a transform function
        and then calls the estimator with predict on transformed data.
        It returns the predictions made.

        In case the last step is no estimator, it returns the transformed data.

        Parameters:
            X:
                Test samples.

            y:
                True values for `X`.

            kwargs:
                Passed to final_estimator.score(), e.g. sample_weight possible.

        Returns:
            Score value.
        """

        X, y, kwargs = self.transform(X, y=y, **kwargs)

        # call score on final estimator
        if self._final_estimator is not None:
            if self._final_estimator.is_estimator:
                return self._final_estimator.score(X, y, **kwargs)

        msg = "It is not possible to run the score method without matching final_estimator. " \
              "Make sure that the last element is an estimator with integrated score function."
        logger.error(msg)
        raise ValueError(msg)

    @property
    def named_steps(self):
        return dict(self.elements)

    @property
    def _final_estimator(self):
        return self.elements[-1][1]

    @property
    def _estimator_type(self):
        return getattr(self._final_estimator, '_estimator_type')

    def clear_cache(self):
        if self.cache_man is not None:
            self.cache_man.clear_cache()

    def add_preprocessing(self, preprocessing):
        if preprocessing:
            self.elements.insert(0, (preprocessing.name, preprocessing))

    @property
    def feature_importances_(self):
        return self.elements[-1][1].feature_importances_



from sklearn.base import BaseEstimator


class CustomYTransformer(BaseEstimator):
    """
    Any algorithm that uses the target values to change the data (e.g. transform male and female differently?)
    or that changes the dataset including targets in total (e.g. sample pairing oversampling)
    """

    def __init__(self, param1=0, param2=None):
        # it is important that you name your params the same in the constructor
        #  stub as well as in your class variables!
        self.param1 = param1
        self.param2 = param2

        self.needs_y = True

    def fit(self, data, targets=None, **kwargs):
        """
        Adjust the underlying model or method to the data.

        Returns
        -------
        IMPORTANT: must return self!
        """
        return self

    def transform(self, data, targets=None, **kwargs):
        """
        Apply the method's logic to the data and targets.
        IMPORTANT: MUST return data AND targets
        """
        return data, targets


from photonai.optimization.base_optimizer import PhotonSlaveOptimizer
from photonai.base.photon_elements import Switch

from photonai.optimization import GridSearchOptimizer, RandomGridSearchOptimizer, \
    SkOptOptimizer, RandomSearchOptimizer


class MetaHPOptimizer(PhotonSlaveOptimizer):
    """
     Searches for the best configuration by randomly testing k possible hyperparameter combinations without grid.
    """

    def __init__(self, **kwargs):

        # get name of optimizer from params
        if not "name" in kwargs:
            raise ValueError("MetaHPOptimizer: Please specify name of optimizer class, "
                             "e.g. random_grid_search with key 'name'")
        self.optimizer_name = kwargs['name']
        if self.optimizer_name == 'smac':
            raise ValueError("MetaHPOptimizer: Using smac does "
                             "not make sense here, as it is by default enabled to handle a switch")
        del kwargs['name']
        self.optimizer_params = kwargs

        self.pipeline_elements = None
        self.current_optimizer = None
        self.n_configurations = -1
        self.estimator_dict = {}
        self.switch_name = ''
        self.switch_estimator_config_key = ''
        self.ask = self.next_config_generator()

        self.OPTIMIZER_DICTIONARY = {'grid_search': GridSearchOptimizer,
                                     'random_grid_search': RandomGridSearchOptimizer,
                                     'sk_opt': SkOptOptimizer,
                                     'random_search': RandomSearchOptimizer}

    def prepare(self, pipeline_elements, maximize_metric):
        if not pipeline_elements:
            raise ValueError("Optimizer cannot optimize empty pipeline")

        if not isinstance((pipeline_elements[-1]), Switch):
            raise ValueError("Cannot use switch_optimizer without Switch Element at the end of pipeline.")

        self.pipeline_elements = pipeline_elements
        switch = self.pipeline_elements[-1]
        self.switch_name = switch.name
        self.switch_estimator_config_key = switch.name + "__estimator_name"
        if self.optimizer_name not in self.OPTIMIZER_DICTIONARY:
            raise ValueError("SwitchOptimizer could not find optimizer {}".format(self.optimizer_name))
        optimizer_class = self.OPTIMIZER_DICTIONARY[self.optimizer_name]

        for element in switch.elements:
            optimizer = optimizer_class(**self.optimizer_params)
            # sub_pipeline = pipeline without estimator plus current estimator
            sub_pipeline = [e for e in self.pipeline_elements[:-1]]
            sub_pipeline.append(element)
            optimizer.prepare(sub_pipeline, maximize_metric)
            self.estimator_dict[element.name] = optimizer
        self.ask = self.next_config_generator()

        if "n_configurations" in self.optimizer_params:
            self.n_configurations = len(switch.elements) * self.optimizer_params["n_configurations"]

    def next_config_generator(self):
        for element_name, optimizer in self.estimator_dict.items():
            self.current_optimizer = optimizer
            for config in self.current_optimizer.ask:
                # update key_names
                config_copy = dict()
                for c_key, c_value in config.items():
                    if c_key.startswith(element_name):
                        config_copy[self.switch_name + "__" + c_key] = c_value
                    else:
                        config_copy[c_key] = c_value
                # append the name of the current estimator to config so we can use it to filter the result by estimator
                config_copy[self.switch_estimator_config_key] = element_name
                yield config_copy

    def tell(self, config, performance):
        # influence return value of next_config
        # remove current estimator name as it is just a hack to filter configs afterwards, and not part of the HP space
        config = dict(config)
        del config[self.switch_estimator_config_key]
        config_copy = dict()
        for c_key, c_value in config.items():
            if c_key.startswith(self.switch_name):
                new_key = c_key.replace(self.switch_name + "__", '')
                config_copy[new_key] = c_value
            else:
                config_copy[c_key] = c_value
        self.current_optimizer.tell(config_copy, performance)



from sklearn.base import BaseEstimator, RegressorMixin, TransformerMixin


class DummyYAndCovariatesTransformer(BaseEstimator):

    def __init__(self):
        self.needs_y = True
        self.needs_covariates = True

    def fit(self, X, y, **kwargs):
        pass

    def transform(self, X, y, **kwargs):

        self.X = X
        self.y = y
        self.kwargs =kwargs

        if y is not None:
            y = y + 1
        if len(kwargs) > 0:
            X = X - 1
            kwargs["sample1_edit"] = kwargs["sample1"] + 5
        return X, y, kwargs


class DummyEstimator(BaseEstimator, RegressorMixin):
    def __init__(self):
        pass

    def fit(self, X, y, **kwargs):
        return self

    def predict(self, X, **kwargs):
        return X + 1


class DummyNeedsCovariatesEstimator(BaseEstimator, RegressorMixin):
    def __init__(self):
        self.needs_covariates = True

    def fit(self, X, y, **kwargs):
        return self

    def predict(self, X, **kwargs):
        return X + 1


class DummyTransformer(BaseEstimator, RegressorMixin):
    def __init__(self):
        pass

    def fit(self, X, y, **kwargs):
        return self

    def transform(self, X, **kwargs):
        return X + 1


class DummyNeedsCovariatesTransformer(BaseEstimator, TransformerMixin):
    def __init__(self):
        self.needs_covariates = True

    def fit(self, X, y, **kwargs):
        return self

    def transform(self, X, **kwargs):
        return X + 1, {'covariates': kwargs['covariates'] + 1}


class DummyNeedsYTransformer(BaseEstimator, TransformerMixin):
    def __init__(self):
        self.needs_y = True

    def fit(self, X, y, **kwargs):
        return self

    def transform(self, X, y):
        return X + 1, y + 1


class DummyNeedsCovariatesAndYTransformer(BaseEstimator, TransformerMixin):
    def __init__(self):
        self.needs_y = True
        self.needs_covariates = True

    def fit(self, X, y, **kwargs):
        return self

    def transform(self, X, y, **kwargs):
        return X + 1, y + 1, {'covariates': kwargs['covariates'] + 1}


class DummyEstimatorWrongType(BaseEstimator):
    _estimator_type = 'clusterer'

    def __init__(self):
        pass

    def fit(self, X, y, **kwargs):
        return self

    def predict(self, X, **kwargs):
        return X


class DummyTransformerWithPredict(BaseEstimator, TransformerMixin):
    def __init__(self):
        pass

    def fit(self, X, y, **kwargs):
        return self

    def predict(self, X, **kwargs):
        return X


class DummyEstimatorNoPredict(BaseEstimator, RegressorMixin):
    def __init__(self):
        pass

    def fit(self, X, y, **kwargs):
        return self





from collections.abc import Iterable
from prettytable import PrettyTable
from photonai.photonlogger.logger import logger
from sklearn.base import BaseEstimator, ClassifierMixin
import numpy as np


class XPredictor(BaseEstimator, ClassifierMixin):

    _estimator_type = 'classifier'

    def __init__(self, change_predictions = False):
        self.needs_y = False
        self.needs_covariates = False
        self.change_predictions = change_predictions
        pass

    def fit(self, X, y=None, **kwargs):
        return self

    def predict(self, X, **kwargs):
        if self.change_predictions:
            # change it relative to value so that it is fold-specific
            return XPredictor.adapt_X(X)
        return X

    @staticmethod
    def adapt_X(X):
        return [i-(0.1*i) for i in X]

    def predict_proba(self, X):
        return X/10


class PhotonPrintHelper:

    @staticmethod
    def config_to_human_readable_dict(pipe, specific_config):
        prettified_config = {}
        for el_key, el_value in specific_config.items():
            items = el_key.split('__')
            name = items[0]
            rest = '__'.join(items[1::])
            if name in pipe.named_steps:
                if not name in prettified_config:
                    prettified_config[name] = list()
                prettified_config[name].append(pipe.named_steps[name].prettify_config_output(rest, el_value))
            else:
                raise ValueError('Item is not contained in pipeline:' + name)
        return prettified_config


class PhotonDataHelper:

    @staticmethod
    def chunker(nr_items, size):
        return [(pos, pos + size) for pos in range(0, nr_items, size)]

    @staticmethod
    def find_n(X):
        if hasattr(X, 'shape'):
            n = X.shape[0]
        elif isinstance(X, Iterable):
            n = len(X)
        else:
            n = 1
        return n

    @staticmethod
    def split_data(X, y, kwargs, start=0, stop=1, indices=None):
        # iterate through data batchwise

        # is that necessary?
        # if isinstance(X, np.ndarray):
        #     dim = len(X.shape)
        # else:
        #     dim = 1
        #
        # if dim > 1:
        #     X_batched = X[start:stop, :]
        # else:
        #     X_batched = X[start:stop]
        # ----------------------------------
        # if we want only one item, we need to make the stop value the excluded upper limit
        if start == stop:
            stop = start + 1

        if indices is None:
            indices_to_use = slice(start, stop)
        else:
            indices_to_use = indices

        if not isinstance(X, np.ndarray):
            X = np.asarray(X)
        X_batched = X[indices_to_use]

        # if we are to batch then apply it
        if y is not None:
            if not isinstance(y, np.ndarray):
                y = np.asarray(y)
            y_batched = y[indices_to_use]
        else:
            y_batched = None

        kwargs_dict_batched = dict()
        if kwargs is not None:
            for key, kwargs_list in kwargs.items():
                if not isinstance(kwargs_list, np.ndarray):
                    kwargs_list = np.array(kwargs_list)
                if len(kwargs_list.shape) > 1:
                    kwargs_dict_batched[key] = kwargs_list[indices_to_use, :]
                else:
                    kwargs_dict_batched[key] = kwargs_list[indices_to_use]

        return X_batched, y_batched, kwargs_dict_batched

    @staticmethod
    def join_data(X, X_new, y, y_new, kwargs, kwargs_new):
        processed_X = PhotonDataHelper.stack_data_vertically(X, X_new)

        if y_new is not None:
            processed_y = PhotonDataHelper.stack_data_vertically(y, y_new)
        else:
            processed_y = None

        processed_kwargs = PhotonDataHelper.join_dictionaries(kwargs, kwargs_new)
        return processed_X, processed_y, processed_kwargs

    @staticmethod
    def join_dictionaries(dict_a: dict, dict_b: dict):
        new_dict = dict()

        if dict_a is None:
            dict_a = dict()
        if dict_b is not None:
            for key, value in dict_b.items():
                if key not in dict_a:
                    new_dict[key] = dict_b[key]
                else:
                    new_dict[key] = PhotonDataHelper.stack_data_vertically(dict_a[key], value)
        return new_dict

    @staticmethod
    def index_dict(d: dict, boolean_index):
        new_dict = dict()
        for key, value in d.items():
            new_dict[key] = value[boolean_index]
        return new_dict

    @staticmethod
    def stack_data_vertically(existing_array, new_array):
        if existing_array is not None and len(existing_array) != 0:
            if isinstance(new_array, np.ndarray) and len(new_array.shape) < 2:
                existing_array = np.hstack((existing_array, new_array))
            elif isinstance(new_array, list):
                existing_array = existing_array + new_array
            elif new_array is None and len(existing_array) == 0:
                return None
            else:
                existing_array = np.vstack((existing_array, new_array))
        else:
            existing_array = new_array
        return existing_array

    @staticmethod
    def stack_data_horizontally(existing_array, new_array):
        """
        Helper method to horizontally join the outcome of each child

        Parameters
        ----------
        * `a` [ndarray]:
            The existing matrix
        * `b` [ndarray]:
            The matrix that is to be attached horizontally

        Returns
        -------
        New matrix, that is a and b horizontally joined

        """
        if existing_array is None or (isinstance(existing_array, np.ndarray) and existing_array.size == 0):
            return new_array
        else:
            # Todo: check for right dimensions!
            if existing_array.ndim == 1 and new_array.ndim == 1:
                existing_array = np.column_stack((existing_array, new_array))
            else:
                if new_array.ndim == 1:
                    new_array = np.reshape(new_array, (new_array.shape[0], 1))
                if existing_array.ndim == 1:
                    existing_array = np.reshape(existing_array, (existing_array.shape[0], 1))
                existing_array = np.concatenate((existing_array, new_array), axis=1)
        return existing_array

    @staticmethod
    def resort_splitted_data(X, y, kwargs, idx_list):
        _sort_order = np.argsort(idx_list)
        X = np.asarray(X)[_sort_order]
        if y is not None:
            y = np.asarray(y)[_sort_order]
        if kwargs is not None:
            for k, v in kwargs.items():
                kwargs[k] = v[_sort_order]
        return X, y, kwargs


def print_metrics(header, metric_dict, summary=False):
    t = PrettyTable(['PERFORMANCE ' + header, ''])
    for m_key, m_value in metric_dict.items():
        t.add_row([m_key, "%.4f" % m_value])
    if summary:
        return t
    logger.photon_system_log(t)


def print_double_metrics(metric_dict_train, metric_dict_test, photon_system_log=True, summary=False):
    t = PrettyTable(['METRIC', 'PERFORMANCE TRAIN', 'PERFORMANCE TEST'])
    for m_key, m_value in metric_dict_train.items():
        t.add_row([m_key, "%.4f" % m_value, "%.4f" % metric_dict_test[m_key]])
    if summary:
        return t
    if photon_system_log:
        logger.photon_system_log(t)
    else:
        logger.debug(t)


def print_estimator_metrics(estimator_performances, metric_list, summary=False):
    t = PrettyTable(['Estimator'] + metric_list)
    for estimator_name, estimator_values in estimator_performances.items():
        t.add_row([estimator_name] + ["%.4f" % estimator_values[m] for m in metric_list])
    if summary:
        return t
    logger.system_line()
    logger.photon_system_log("AVERAGE PERFORMANCE PER ESTIMATOR")
    logger.system_line()
    logger.photon_system_log(t)


def print_config_list_table(estimator_name="", list_of_config_dicts=list(), summary=False):

    parameter_names = list(set().union(*list_of_config_dicts))
    t = PrettyTable(['Config #'] + parameter_names)
    for idx, config in enumerate(list_of_config_dicts):
        t.add_row([str(idx)] + [config[p] for p in parameter_names])
    if summary:
        return t

    logger.system_line()
    logger.photon_system_log("Estimator Name")
    logger.system_line()
    logger.photon_system_log(t)



import unittest
import numpy as np
import warnings
from sklearn.base import BaseEstimator

from photonai.base import PipelineElement


class DummyBatchTransformer(BaseEstimator):

    def __init__(self):
        self.needs_y = True
        self.needs_covariates = True
        self.predict_count = 0

    def fit(self, X, y, **kwargs):
        pass

    def transform(self, X, y=None, **kwargs):
        if hasattr(X, '__iter__'):
            X_new = []
            for i, x in enumerate(X):
                X_new.append([str(sub_x) + str(y[i]) for sub_x in x])

            if not np.array_equal(y, kwargs["animals"]):
                raise Exception("Batching y and kwargs delivery is strange")

            kwargs["animals"] = [i[::-1] for i in kwargs["animals"]]

            return np.asarray(X_new), np.asarray(y), kwargs
        else:
            return X, y, kwargs

    def predict(self, X, y=None, **kwargs):
        if hasattr(X, '__iter__'):
            self.predict_count += 1
            if not hasattr(X, 'shape'):
                return self.predict_count
            predictions = np.ones((X.shape[0],)) * self.predict_count
            return predictions
        else:
            return y


class BatchingTests(unittest.TestCase):

    def setUp(self):
        self.batch_size = 10
        nr_features = 3
        origin_list = ["affe", "tiger", "schwein", "giraffe", "löwe"]
        self.data = None
        self.targets = None

        self.neuro_batch = PipelineElement("dummy_batch", batch_size=self.batch_size,
                                           base_element=DummyBatchTransformer())

        for element in origin_list:
            features = [element + str(i) for i in range(0, nr_features)]
            if self.data is None:
                self.data = np.array([features] * self.batch_size)
            else:
                self.data = np.vstack((self.data, [features] * self.batch_size))
            if self.targets is None:
                self.targets = np.array([element] * self.batch_size)
            else:
                self.targets = np.hstack((self.targets,  [element] * self.batch_size))

        self.data = np.array(self.data)
        self.targets = np.array(self.targets)
        self.kwargs = {"animals": self.targets}

    def test_transform(self):
        X_new, y_new, kwargs_new = self.neuro_batch.transform(self.data, self.targets, **self.kwargs)
        self.assertListEqual(X_new[0, :].tolist(), ["affe0affe", "affe1affe", "affe2affe"])
        self.assertListEqual(X_new[49, :].tolist(), ["löwe0löwe", "löwe1löwe", "löwe2löwe"])
        self.assertEqual(kwargs_new["animals"][0], "effa")
        self.assertEqual(kwargs_new["animals"][49], "ewöl")

        with warnings.catch_warnings(record=True) as w:
            self.neuro_batch.transform('str')
            assert any("Cannot do batching" in s for s in [e.message.args[0] for e in w])

    def test_predict(self):
        y_predicted = self.neuro_batch.predict(self.data, **self.kwargs)
        # assure that predict is batch wisely called
        self.assertEqual(y_predicted[0], 1)
        self.assertEqual(y_predicted[-1], (self.data.shape[0]/self.batch_size))
        with warnings.catch_warnings(record=True) as w:
            self.neuro_batch.predict('str')
            assert any("Cannot do batching" in s for s in [e.message.args[0] for e in w])



from numpy.testing import assert_array_almost_equal
from sklearn.datasets import load_breast_cancer, load_boston
from sklearn.model_selection import KFold, ShuffleSplit

from photonai.base import Hyperpipe, PipelineElement
from photonai.helper.photon_base_test import PhotonBaseTest


class FeatureSelectionTests(PhotonBaseTest):

    @classmethod
    def setUpClass(cls) -> None:
        cls.file = __file__
        super(FeatureSelectionTests, cls).setUpClass()

    def setUp(self):
        super(FeatureSelectionTests, self).setUp()
        self.X_classif, self.y_classif = load_breast_cancer(return_X_y=True)
        self.X_regr, self.y_regr = load_boston(return_X_y=True)
        self.pipe_classif = Hyperpipe("feature_selection_pipe_classif",
                              outer_cv=ShuffleSplit(test_size=0.2, n_splits=1, random_state=15),
                              inner_cv= KFold(n_splits=3, shuffle=True, random_state=15),
                              metrics=["accuracy"], best_config_metric="accuracy",
                              project_folder=self.tmp_folder_path)
        self.pipe_regr = Hyperpipe("feature_selection_pipe_regr",
                              outer_cv=ShuffleSplit(test_size=0.2, n_splits=1, random_state=15),
                              inner_cv= KFold(n_splits=3, shuffle=True, random_state=15),
                              metrics=["mean_absolute_error"], best_config_metric="mean_absolute_error",
                              project_folder=self.tmp_folder_path)

    def test_FRegressionFilterPValue(self):
        self.pipe_regr += PipelineElement('FRegressionFilterPValue')
        self.pipe_regr += PipelineElement('SVR')
        self.pipe_regr.fit(self.X_regr, self.y_regr)

    def test_FRegressionFilterPValue_inverse(self):
        frfpv = PipelineElement('FRegressionFilterPValue', p_threshold=0.001)
        frfpv.fit(self.X_regr[:30], self.y_regr[:30])
        X_selected, _, _ = frfpv.transform(self.X_regr)
        X_back, _, _ = frfpv.inverse_transform(X_selected)
        self.assertLess(X_selected.shape[1], X_back.shape[1])
        self.assertEqual(self.X_regr.shape, X_back.shape)
        assert_array_almost_equal(X_selected,  frfpv.transform(X_back)[0])

        with self.assertRaises(ValueError):
            frfpv.inverse_transform(X_selected[:, :int(X_selected.shape[1]*0.5)])

    def test_FRegressionSelectPercentile(self):
        self.pipe_regr += PipelineElement('FRegressionSelectPercentile')
        self.pipe_regr += PipelineElement('SVR')
        self.pipe_regr.fit(self.X_regr, self.y_regr)

    def test_FRegressionSelectPercentile_inverse(self):
        frsp = PipelineElement('FRegressionSelectPercentile', percentile=5)
        frsp.fit(self.X_regr[:30], self.y_regr[:30])
        X_selected, _, _ = frsp.transform(self.X_regr)
        X_back, _, _ = frsp.inverse_transform(X_selected)
        self.assertLess(X_selected.shape[1], X_back.shape[1])
        self.assertEqual(self.X_regr.shape, X_back.shape)
        assert_array_almost_equal(X_selected,  frsp.transform(X_back)[0])

    def test_FClassifSelectPercentile(self):
        self.pipe_classif += PipelineElement('FClassifSelectPercentile')
        self.pipe_classif += PipelineElement('SVC')
        self.pipe_classif.fit(self.X_classif, self.y_classif)

    def test_FClassifSelectPercentile_inverse(self):
        fcsp = PipelineElement('FClassifSelectPercentile', percentile=5)
        fcsp.fit(self.X_classif[:30], self.y_classif[:30])
        X_selected, _, _ = fcsp.transform(self.X_classif)
        X_back, _, _ = fcsp.inverse_transform(X_selected)
        self.assertLess(X_selected.shape[1], X_back.shape[1])
        self.assertEqual(self.X_classif.shape, X_back.shape)
        assert_array_almost_equal(X_selected,  fcsp.transform(X_back)[0])

    def test_LassoFeatureSelection(self):
        self.pipe_regr += PipelineElement('LassoFeatureSelection')
        self.pipe_regr += PipelineElement('SVR')
        self.pipe_regr.fit(self.X_regr, self.y_regr)

    def test_LassoFeatureSelection_inverse(self):
        lfs = PipelineElement('LassoFeatureSelection')
        lfs.fit(self.X_regr[:30], self.y_regr[:30])
        X_selected, _, _ = lfs.transform(self.X_regr)
        X_back, _, _ = lfs.inverse_transform(X_selected)
        self.assertLess(X_selected.shape[1], self.X_regr.shape[1])
        self.assertEqual(self.X_regr.shape, X_back.shape)
        assert_array_almost_equal(X_selected,  lfs.transform(X_back)[0])

        with self.assertRaises(ValueError):
            lfs.inverse_transform(X_selected[:, :int(X_selected.shape[1]*0.5)])



from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
import numpy as np
from scipy import stats


class PhotonVotingClassifier(BaseEstimator, ClassifierMixin):

    def __init__(self):

        self.strategy = PhotonVotingClassifier._most_frequent

    def fit(self, X, y=None, **kwargs):
        return self

    def predict(self, X, **kwargs):
        if X is not None:
            output = self.strategy(X, axis=1)
            if isinstance(X[0], int):
                return [np.round(output)]
            else:
                return output

    @staticmethod
    def _most_frequent(X, axis):
        mode_obj = stats.mode(X, axis=axis)
        return [i[0] for i in mode_obj.mode]


class PhotonVotingRegressor(BaseEstimator, RegressorMixin):

    def __init__(self, strategy='mean'):

        self.STRATEGY_DICT = {'mean': np.mean,
                              'median': np.median}

        self._strategy = None
        self.strategy = strategy

    @property
    def strategy(self):
        return self._strategy

    @strategy.setter
    def strategy(self, strategy_val):
        if strategy_val not in self.STRATEGY_DICT:
            raise ValueError("Strategy " + str(self.strategy) + " is not supported right now. ")
        else:
            self._strategy = strategy_val

    def fit(self, X, y=None, **kwargs):
        return self

    def predict(self, X, **kwargs):
        if X is not None:
            output = self.STRATEGY_DICT[self.strategy](X, axis=1)
            if isinstance(X[0], int):
                return [np.round(output)]
            else:
                return output



from typing import Callable


class PhotonSlaveOptimizer(object):
    """Photon slave optimizer.

    Base class for optimizer in PHOTONAI.
    The PhotonSlaveOptimizer is controlled by PHOTONAIs OuterFoldManager.
    With the ask-tell principle PHOTONAI gets new configs.
    It terminates by some specific criteria, that leads to an empty ask generator.

    """
    def prepare(self, pipeline_elements: list, maximize_metric: bool) -> None:
        """
        Initializes hyperparameter search.

        Assembles all hyperparameters of the list of PipelineElements
        in order to prepare the hyperparameter search space.
        Hyperparameters can be accessed via pipe_element.hyperparameters.

        Parameters:
            pipeline_elements:
                List of all PipelineElements to create the hyperparameter space.

            maximize_metric:
                Boolean to distinguish between score and error.

        """
        pass

    def ask(self) -> dict:
        """
        When ask is called it returns the next configuration to be tested.

        Returns:
            Dict of configurations with parameters to be tested next.

        """
        pass

    def tell(self, config: dict, performance: float) -> None:
        """
        Returns the performance of a tested configuration to calculate new ones.

        Parameters:
            config: dict
                The configuration that has been trained and tested.

            performance: float
                Metrics about the configuration's generalization capabilities.

        """
        pass


class PhotonMasterOptimizer(object):
    """Photon master optimizer.

    Base class for optimizer in PHOTONAI.
    The PhotonMasterOptimizer controls PHOTONAIs optimization process.
    PHOTONAI creates an objective function that is used by the optimizer.
    The limitation of runs and configs of the
    objective function is up to PhotonMasterOptimizer.

    """
    def prepare(self, pipeline_elements: list, maximize_metric: bool, objective_function: Callable) -> None:
        """
        Initializes hyperparameter search.

        Assembles all hyperparameters of the list of PipelineElements
        in order to prepare the hyperparameter search space.
        Hyperparameters can be accessed via pipe_element.hyperparameters.

        Parameters:
            pipeline_elements:
                List of all PipelineElements to create the hyperparameter space.

            maximize_metric:
                Boolean to distinguish between score and error.

            objective_function:
                The cost or objective function.

        """
        pass

    def optimize(self) -> None:
        """Start the optimization process based on the underlying objective function."""
        pass



import warnings
import numpy as np
import keras
from typing import Union
from keras.utils import to_categorical
from keras.layers import Dropout, Dense
from keras.layers.normalization import BatchNormalization
from keras.models import Sequential
from keras.optimizers import Optimizer, Adam, RMSprop, Adadelta, Adagrad, Adamax, Nadam, SGD
from keras.activations import softmax, softplus, selu, sigmoid, softsign, hard_sigmoid, elu, relu, tanh, \
    linear, exponential
from sklearn.base import ClassifierMixin, RegressorMixin

from photonai.photonlogger.logger import logger
from photonai.modelwrapper.keras_base_estimator import KerasBaseEstimator

__supported_optimizers__ = {
    'sgd': SGD,
    'rmsprop': RMSprop,
    'adagrad': Adagrad,
    'adadelta': Adadelta,
    'adam': Adam,
    'adamax': Adamax,
    'nadam': Nadam
}
__supported_activations__ = {
    'softmax': softmax,
    'softplus': softplus,
    'selu': selu,
    'sigmoid': sigmoid,
    'softsign': softsign,
    'hard_sigmoid': hard_sigmoid,
    'elu': elu,
    'relu': relu,
    'tanh': tanh,
    'linear': linear,
    'exponential': exponential
}

__allocation_loss_functions__ = {
    'regression': ['mean_squared_error', 'mean_squared_logarithmic_error', 'mean_absolute_error'],
    'binary_classification': ['binary_crossentropy', 'hinge', 'squared_hinge'],
    'multi_classification': ['categorical_crossentropy', 'sparse_categorical_crossentropy',
                             'kullback_leibler_divergence']
}


class KerasBaseClassifier(KerasBaseEstimator, ClassifierMixin):

    def __init__(self,
                 model=None,
                 epochs: int = 10,
                 callbacks: list = None,
                 validation_split: float = 0.1,
                 nn_batch_size: int = 33,
                 multi_class: bool = True,
                 verbosity: int = 0):
        super(KerasBaseClassifier, self).__init__(model=model,
                                                  epochs=epochs,
                                                  callbacks=callbacks,
                                                  validation_split=validation_split,
                                                  nn_batch_size=nn_batch_size,
                                                  verbosity=verbosity)
        self.multi_class = multi_class

    def predict(self, X):
        predict_result = self.model.predict(X, batch_size=self.nn_batch_size)
        if self.multi_class:
            max_index = np.argmax(predict_result, axis=1)
        else:
            max_index = np.array([val > 0.5 for val in predict_result])
        return max_index

    def encode_targets(self, y):
        if self.multi_class:
            return to_categorical(y)
        else:
            return y


class KerasBaseRegressor(KerasBaseEstimator, RegressorMixin):

    def __init__(self,
                 model=None,
                 epochs: int = 10,
                 validation_split: float = 0.1,
                 nn_batch_size: int = 64,
                 callbacks: list = None,
                 verbosity: int = 0):
        super(KerasBaseRegressor, self).__init__(model=model,
                                                 epochs=epochs,
                                                 validation_split=validation_split,
                                                 nn_batch_size=nn_batch_size,
                                                 callbacks=callbacks,
                                                 verbosity=verbosity)

    def predict(self, X):
        return np.array([val[0] for val in self.predict_proba(X)])


class KerasDnnBaseModel(KerasBaseEstimator):

    def __init__(self, hidden_layer_sizes: list = None,
                 target_dimension: int = 2,
                 target_activation: str = "softmax",
                 learning_rate: float = 0.1,
                 loss: str = "categorical_crossentropy",
                 metrics: list = None,
                 batch_normalization: bool = True,
                 verbosity: int = 0,
                 dropout_rate: Union[list, float] = 0.2,
                 activations: Union[list, str] = 'relu',
                 optimizer="adam"  # list or keras.optimizer
                 ):

        self._hidden_layer_sizes = None
        self._dropout_rate = None
        self._target_activation = None
        self._optimizer = None
        self._loss = None
        self._activations = None
        self._metrics = None

        self.hidden_layer_sizes = hidden_layer_sizes
        self.dropout_rate = dropout_rate
        self.target_dimension = target_dimension
        self.target_activation = target_activation
        self.activations = activations
        self.learning_rate = learning_rate
        self.optimizer = optimizer
        self.loss = loss
        self.batch_normalization = batch_normalization
        self.metrics = metrics

        self.verbosity = verbosity

        self.model = None

    @property
    def hidden_layer_sizes(self):
        """
        Getter for attribute dropout.
        :return:
        """
        return self._hidden_layer_sizes

    @hidden_layer_sizes.setter
    def hidden_layer_sizes(self, value):
        """
        Setter for dropout_rate. Checks if strategy is supported.
        :param value:
        :return:
        """
        self._hidden_layer_sizes = value
        if not value:
            return
        layer_count = len(value)
        if isinstance(self.activations, list):
            if len(self.activations) != layer_count:
                raise ValueError("activations length missmatched layer length.")
        else:
            self._activations = [self.activations] * layer_count
        if isinstance(self.dropout_rate, list):
            if len(self.dropout_rate) != layer_count:
                raise ValueError("dropout_rate length missmatched layer length.")
        else:
            self._dropout_rate = [self.dropout_rate] * layer_count

    @property
    def dropout_rate(self):
        """
        Getter for attribute dropout.
        :return:
        """
        return self._dropout_rate

    @dropout_rate.setter
    def dropout_rate(self, value):
        """
        Setter for dropout_rate. Checks if strategy is supported.
        :param value:
        :return:
        """
        if not type(value) in [list, float]:
            raise ValueError("Dropout type is not supported. Please use one of [list, float]")
        else:
            if not self._hidden_layer_sizes:
                self._dropout_rate = value
            else:
                if type(value) == float:
                    self._dropout_rate = [value]*len(self.hidden_layer_sizes)
                    msg = "Dropout with type float converted to type list."
                    logger.warning(msg)
                    warnings.warn(msg)
                elif len(value) != len(self.hidden_layer_sizes):
                    raise ValueError("Dropout length missmatched layer length.")
                else:
                    self._dropout_rate = value

    @property
    def activations(self):
        """
        Getter for attribute act_func.
        :return:
        """
        return self._activations

    @activations.setter
    def activations(self, value):
        """
        Setter for act_func. Checks if strategy is supported.
        :param value:
        :return:
        """
        if not type(value) in [list, str]:
            raise ValueError("act_func type is not supported. Please use one of [list, float]")
        else:
            if not self._hidden_layer_sizes:
                self._activations = value
            else:
                if type(value) == str:
                    if value in __supported_activations__.keys():
                        self._activations = [value]*len(self.hidden_layer_sizes)
                        msg = "activations with type str converted to type list."
                        logger.warning(msg)
                        warnings.warn(msg)
                    else:
                        raise ValueError(
                            "activations not supported. Please use one of: " + str(__supported_activations__.keys()))
                elif len(value) != len(self.hidden_layer_sizes):
                    raise ValueError("activations length missmatched layer length.")
                elif any(act not in __supported_activations__.keys() for act in value):
                    raise ValueError("activations not supported. Please use one of: "+str(__supported_activations__.keys()))
                else:
                    self._activations = value

    @property
    def optimizer(self):
        return self._optimizer

    @optimizer.setter
    def optimizer(self, value):
        if isinstance(value, Optimizer):
            self._optimizer = value
        if value.lower() not in __supported_optimizers__.keys():
            raise ValueError("Optimizer is not supported by keras. Please use one of: "+str(__supported_optimizers__))
        else:
            self._optimizer = __supported_optimizers__[value.lower()](lr=self.learning_rate)

    @property
    def target_activation(self):
        return self._target_activation

    @target_activation.setter
    def target_activation(self, value):
        if value in __supported_activations__.keys():
            self._target_activation = value
        else:
            raise ValueError("target_activation not supported. Please use one of: " + str(__supported_activations__.keys()))

    @property
    def loss(self):
        return self._loss

    @loss.setter
    def loss(self, value):
        try:
            keras.losses.get(value)
            self._loss = value
        except ValueError:
            raise ValueError("Unknown loss function:" + value)

    @property
    def metrics(self):
        return self._metrics

    @metrics.setter
    def metrics(self, value):
        # TODO: check metrics values!
#        try:
#            all(keras.metrics.get(metric) for metric in value)
            self._metrics = value
#        except ValueError:
#            raise ValueError("Unknown metric function in:" + str(value))

    def create_model(self, input_size):

        self.model = Sequential()
        for i, size in enumerate(self.hidden_layer_sizes):
            if i == 0:
                self.model.add(Dense(size, input_dim=input_size, activation=self.activations[i]))
            else:
                self.model.add(Dense(size, activation=self.activations[i]))
            self.model.add(Dropout(rate=self.dropout_rate[i]))

            if self.batch_normalization == 1:
                self.model.add(BatchNormalization())

        self.model.add(Dense(self.target_dimension, activation=self.target_activation))

        # Compile model
        self.model.compile(loss=self.loss, optimizer=self.optimizer, metrics=self.metrics)

        self.init_weights = self.model.get_weights()

        return self


def get_loss_allocation():
    return __allocation_loss_functions__



"""
PHOTONAI
is a rapid prototyping framework enabling (not so experienced) users to build, train, optimize, evaluate,
and share even complex machine learning pipelines with very high efficiency.

PHOTONAI is an object-oriented python framework designed to leave you deciding the important things and automatizing the rest:
 By treating each pipeline element as a building block, we create a system in which the user can select and combine
 processing steps, adapt their arrangement or stack them in more advanced pipeline layouts.

PHOTONAI gives you an easy way of setting up a full stack machine learning pipeline including nested cross-validation and hyperparameter search.
After PHOTON has found the best configuration for your model, it offers a convenient possibility to explore the analyzed hyperparameter space.
It also enables you to persist and load your optimal model, including all preprocessing elements, with only one line of code.

"""

__version__ = '2.1.0'

# __all__ = ( )



from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import KFold

from photonai.base import Hyperpipe, PipelineElement
from photonai.optimization import IntegerRange

# WE USE THE BREAST CANCER SET FROM SKLEARN
X, y = load_breast_cancer(return_X_y=True)

# DESIGN YOUR PIPELINE
my_pipe = Hyperpipe('multi_perceptron_pipe',
                    optimizer='sk_opt',
                    optimizer_params={'n_configurations': 25},
                    metrics=['accuracy', 'precision', 'recall', 'balanced_accuracy'],
                    best_config_metric='accuracy',
                    outer_cv=KFold(n_splits=3),
                    inner_cv=KFold(n_splits=3),
                    verbosity=1,
                    project_folder='./tmp/')


# ADD ELEMENTS TO YOUR PIPELINE
my_pipe += PipelineElement('StandardScaler')

my_pipe += PipelineElement('PhotonMLPClassifier', hyperparameters={'layer_1': IntegerRange(1, 5),
                                                                   'layer_2': IntegerRange(0, 5),
                                                                   'layer_3': IntegerRange(0, 5)})

# NOW TRAIN YOUR PIPELINE
my_pipe.fit(X, y)
